"""
nc -4 -u p2pd.net 7

"""

from p2pd.test_init import *
try:
    from .static_route import *
except:
    from static_route import *
import os
from p2pd.utils import log_exception, what_exception
from p2pd import STUNClient, Interface
from p2pd.settings import *
from p2pd.net import VALID_AFS, TCP, UDP
from p2pd.nat import *
from p2pd.base_stream import pipe_open, SUB_ALL, BaseProto
from p2pd.stun_client import tran_info_patterns, do_stun_request
from p2pd.stun_client import changeRequest, changePortRequest
from p2pd.ip_range import IPRange

env = os.environ.copy()
class TestStunClient(unittest.IsolatedAsyncioTestCase):
    async def test_stun_client(self):
        one_valid = False
        for af in VALID_AFS:
            try:
                # Get default Interface for AF type.
                i = Interface(af)
                #rp = use_fixed_rp(i)
                await i.start()
                one_valid = True
            except Exception:
                # Skip test if not supported.
                continue

            try:
                # Test echo server with AF.
                stun_client = STUNClient(i, af)
                wan_ip = await stun_client.get_wan_ip()
                self.assertTrue(wan_ip)

                m = await stun_client.get_mapping(proto=TCP) 
                self.assertTrue(isinstance(m[0], Interface)) # Interface used for stun.
                self.assertTrue(isinstance(m[1], BaseProto))  # Instance of open socket to stun server.
                self.assertTrue(isinstance(m[2], int)) # Local port
                self.assertTrue(isinstance(m[3], int)) # Remote mapping
                self.assertTrue(isinstance(m[4], str)) # Remote IP
                self.assertTrue(m[2]) # Local port
                self.assertTrue(m[3]) # Mapped port
                IPRange(m[4]) # Is valid IP

                # Stun server addr.
                route = await i.route(af).bind()
                stun_server = STUNT_SERVERS[af][0]
                dest = await Address(
                    stun_server["primary"]["ip"],
                    stun_server["primary"]["port"],
                    route,
                    UDP
                ).res()

                # Check NAT test result is as expected.
                # Then check that other STUN requests work.
                if env.get("NAT_OPEN_INTERNET", False):
                    nat_type = await stun_client.get_nat_type()
                    if nat_type not in [OPEN_INTERNET, SYMMETRIC_UDP_FIREWALL]:
                        continue

                    for req in [changeRequest, changePortRequest]:
                        # Test change port request.
                        pipe = (await pipe_open(
                            UDP,
                            route,
                            dest
                        )).subscribe(SUB_ALL)

                        # Used for matching the TXID for the stun reply.
                        tran_info = tran_info_patterns(dest.tup)
                        tran_info[1] = 0 # Match any host tuple.
                        extra_data = req
                        pipe.subscribe(tran_info[:2])

                        # Do the change request.
                        ret = await do_stun_request(pipe, dest, tran_info, extra_data=extra_data)
                        if nat_type == OPEN_INTERNET:
                            self.assertTrue(ret["resp"])
                        await pipe.close()
            except:
                continue


        self.assertTrue(one_valid)
                

if __name__ == '__main__':
    main()


import sys
import socket
import platform
import struct
import ipaddress
import random
import copy
import ssl
from io import BytesIO
from .errors import *
from .cmd_tools import *

"""
You're supposed to have unique [src ip, src port, dest ip, dest port] 'tuples' for every end point but you can easily bypass this with UDP. But if you end up with multiple 'sockets' bound to the same endpoint then the OS is not going to route packets correctly. It might make sense to detect this somehow if debug mode is on and show a warning if such actions are detected.

There is also the case where you open a second socket with
the same endpoint and fail to track the state of the first
(and properly close it) which will cause messages to be
routed to the first socket and / or second. It will be
very hard to detect. Sockets need to be properly cleaned
up to avoid this error state.
"""

# Address class has determined host input is a domain.
HOST_TYPE_DOMAIN = 0

# Address class has determine host input is an IP.
HOST_TYPE_IP = 1

# Used to signal preferences for an IP family when multiple
# options are available after resolving a domain.
# For IPs passed it serves as addition error checking.
AF_ANY = 1337

# Error value.
AF_NONE = 80085

# Avoid annoying socket... to access vars.
AF_INET = socket.AF_INET
AF_INET6 = socket.AF_INET6
TCP = STREAM = SOCK_STREAM = socket.SOCK_STREAM
UDP = DGRAM = SOCK_DGRAM = socket.SOCK_DGRAM
RUDP = 1234

# Interfaces are categorized as whether they're ethernet or wireless.
INTERFACE_UNKNOWN = 0
INTERFACE_ETHERNET = 1
INTERFACE_WIRELESS = 2

# Network stack couldn't be determined.
UNKNOWN_STACK = 0

# Stack only supports IPv4.
IP4 = V4 = V4_STACK = AF_INET

# Stack only supports IPv6.
IP6 = V6 = V6_STACK = AF_INET6

V6_LINK_LOCAL_MASK = "fe80" + (":0000" * 7)

# Stack supports both IPv4 and IPv6.
DUEL_STACK = AF_ANY

# Valid stack lists.
VALID_AFS = [IP4, IP6]
VALID_STACKS = [DUEL_STACK, IP4, IP6]

# Used as a timeout argument to recv.
# Non_blocking means it will return immediately even when it has no data.
NET_NON_BLOCKING = 0

# This means it will only return when a message is received.
# Not good if the code is in a processing loop.
NET_BLOCKING = None

# Keep around 1000 messages that haven't been processed.
# Packets are dropped after that point.
NET_MAX_MSG_NO = 1000

# Maximum amount in bytes all the messages can add up to.
NET_MAX_MSGS_SIZEOF = 2 * 1024 * 1024

# Netmasks that are for public addresses.
ZERO_NETMASK_IP4 = "0.0.0.0"
ZERO_NETMASK_IP6 = "0000:0000:0000:0000:0000:0000:0000:0000"
BLACK_HOLE_IPS = {
    IP4: "192.0.2.1",
    IP6: "0100:0000:0000:0000:0000:0000:0000:0001"
}

# A value meaning 'listen to' or 'subscribe to' all messages.
VALID_LOOPBACKS = ["127.0.0.1", "::1"]
VALID_ANY_ADDR = ["0.0.0.0", "::"]
ANY_ADDR = ["0.0.0.0", "ff02::1", "::/0", "255.255.255.255"]
NIC_BIND = 1
EXT_BIND = 2
LOOPBACK_BIND = 3
NODE_PORT = 10001

# Address object types.
IPA_TYPES = ipa_types = (ipaddress.IPv4Address, ipaddress.IPv6Address)

# Convert string proto values to enums.
PROTO_LOOKUP = {
    "TCP": TCP,
    "UDP": UDP,
    "RUDP": RUDP
}

DATAGRAM_TYPES = [
    asyncio.selector_events._SelectorDatagramTransport,
    asyncio.DatagramTransport
]
if sys.platform == "win32":
    if hasattr(asyncio.proactor_events, "_ProactorDatagramTransport"):
        DATAGRAM_TYPES.append(asyncio.proactor_events._ProactorDatagramTransport)

STREAM_TYPES = [asyncio.Transport]
if sys.platform == "win32":
    STREAM_TYPES.append(asyncio.proactor_events._ProactorSocketTransport)

DATAGRAM_TYPES = tuple(DATAGRAM_TYPES)
STREAM_TYPES = tuple(STREAM_TYPES)

IP_PRIVATE = 1
IP_PUBLIC = 2
NOT_WINDOWS = platform.system() != "Windows"

# Fine tune various network settings.
NET_CONF = {
    # Wrap socket with SSL.
    "use_ssl": 0,

    # Timeout for SSL handshake.
    "ssl_handshake": 4,

    # Protocol family used for the socket.socket function.
    "sock_proto": 0,

    # N seconds before a registering recv timeout.
    "recv_timeout": 2,

    # Only applies to TCP.
    "con_timeout": 2,

    # No of messages to receive per subscription.
    "max_qsize": 0,

    # Require unique messages or not.
    "enable_msg_ids": 0,

    # Number of message IDs to keep around.
    "max_msg_ids": 1000,

    # Reuse address tuple for bind() socket call.
    "reuse_addr": False,

    # Setup socket as a broadcast socket.
    "broadcast": False,

    # Buf size for asyncio.StreamReader.
    "reader_limit": 2 ** 16,

    # Return the sock instead of the base proto.
    "sock_only": False,

    # Disable closing sock on error.
    "no_close": True,

    # Whether to set SO_LINGER. None = off.
    # Non-none = linger value.
    "linger": None,

    # Ref to an event loop.
    "loop": None
}

af_to_v = lambda af: 4 if af == IP4 else 6
v_to_af = lambda v: IP4 if v == 4 else IP6
max_cidr = lambda af: 32 if af == IP4 else 128

class FakeSocket():
    def __init__(self, response_bytes):
        self._file = BytesIO(response_bytes)

    def makefile(self, *args, **kwargs):
        return self._file

def af_from_ip_s(ip_s):
    ip_s = to_s(ip_s)
    ip_obj = ip_f(ip_s)
    return v_to_af(ip_obj.version)

def generate_mac(uaa=False, multicast=False, oui=None, separator=':', byte_fmt='%02x'):
    def rand_by(n):
        return [random.randrange(256) for _ in range(n)]

    mac = rand_by(8)
    if oui:
        if type(oui) == str:
            oui = [int(chunk) for chunk in oui.split(separator)]
        mac = oui + rand_by(num=6-len(oui))
    else:
        if multicast:
            mac[0] |= 1 # set bit 0
        else:
            mac[0] &= ~1 # clear bit 0
        if uaa:
            mac[0] &= ~(1 << 1) # clear bit 1
        else:
            mac[0] |= 1 << 1 # set bit 1
    return separator.join(byte_fmt % b for b in mac)

def mac_to_b(mac_str):
    mac = mac_str.replace(":", "")
    mac = mac.strip()
    while len(mac) < 12 :
        mac = '0' + mac
    macb = b''
    for i in range(0, 12, 2) :
        m = int(mac[i:i + 2], 16)
        macb += struct.pack('!B', m)
    return macb

async def sock_close_all(sock_list):
    if sock_list is not None:
        # Make it iterable.
        if type(sock_list) != list:
            sock_list = [ sock_list ]

        # Close all socks.
        for sock in sock_list:
            if sock is not None:
                await sock.close()

    return []

def ip_str_to_int(ip_str):
    ip_obj = ipaddress.ip_address(ip_str)
    if ip_obj.version == 4:
        pack_ip = socket.inet_aton(ip_str)
        return struct.unpack("!L", pack_ip)[0]
    else:
        ip_str = str(ip_obj.exploded)
        hex_str = to_h(socket.inet_pton(
            AF_INET6, ip_str
        ))
        return to_i(hex_str)

def netmask_to_cidr(netmask):
    # Already a cidr.
    if "/" in netmask:
        return int(netmask.replace("/", ""))

    as_int = ip_str_to_int(netmask) 
    return bin(as_int).count("1")

def cidr_to_netmask(cidr, af):
    end = 32 if af == AF_INET else 128
    buf = "1" * cidr
    buf += "0" * (end - cidr)
    n = int(buf, 2)
    if af == AF_INET:
        return (str(ipaddress.IPv4Address(n)))
    else:
        return str(ipaddress.IPv6Address(n).exploded)

def toggle_host_bits(netmask, ip_str, toggle=0):
    ip_obj = ipaddress.ip_address(ip_str)
    if "/" in netmask:
        cidr = int(netmask.split("/")[-1])
    else:
        cidr = netmask_to_cidr(netmask)
    as_int = ip_str_to_int(ip_str)
    as_bin = bin(as_int)[2:]
    net_part = as_bin[:cidr]
    if not toggle:
        host_part = "0" * (len(as_bin) - len(net_part))
    else:
        host_part = "1" * (len(as_bin) - len(net_part))

    bin_result = net_part + host_part
    n_result = int(bin_result, 2)
    if ip_obj.version == 4:
        return str(ipaddress.IPv4Address(n_result))
    else:
        return str(ipaddress.IPv6Address(n_result).exploded)

def get_broadcast_ip(netmask, gw_ip):
    return toggle_host_bits(netmask, gw_ip, toggle=1)

"""
Host IP just needs to be any valid host IP in
the given subnet that corrosponds to the netmask.
Then the host portion is manipulated and the
network past is left alone.

Netmask is an IP in the same format as host_IP
with all the bits from the left set to 1 to
mask the network portion. Both netmask and
host_ip can be v4 or v6.

N is the offset from the last valid IP in the range.
Want the last valid IP? Set it to 0. Or the 10th from
the last? Then set it to 9.
"""
def ip_from_last(n, netmask, host_ip):
    assert(n)
    min_ip_str = toggle_host_bits(netmask, host_ip, toggle=0)
    max_ip_str = toggle_host_bits(netmask, host_ip, toggle=1)
    ip_obj = ipaddress.ip_address(host_ip)
    if ip_obj.version == 4:
        min_ip_n = int(ipaddress.IPv4Address(min_ip_str)) + 1
        max_ip_n = int(ipaddress.IPv4Address(max_ip_str)) - 1
    else:
        min_ip_n = int(ipaddress.IPv6Address(min_ip_str)) + 1
        max_ip_n = int(ipaddress.IPv6Address(max_ip_str)) - 1

    ip_n = -n % (max_ip_n + 1)
    if ip_n <= min_ip_n:
        ip_n = (-(min_ip_n - (n or min_ip_n)) or -1) % max_ip_n

    if ip_obj.version == 4:
        return str(ipaddress.IPv4Address(ip_n))
    else:
        return str(ipaddress.IPv6Address(ip_n).exploded)

def ipv6_rand_host(ip_val, cidr):
    # Back to an address.
    host_no = random.randrange(0, 2 ** cidr)
    int_val = int(ipaddress.IPv6Address(ip_val)) + host_no
    ip_str = str(ipaddress.IPv6Address(int_val))

    return ip_norm(ip_str)

def ipv6_rand_link_local():
    # Network portion for IPv6 link-local addresses
    # fe80:0000:0000:0000
    buf = "111111101" + ("0" * 55)

    # Blank host portion placeholder.
    buf += ("0" * 64)
    assert(len(buf) == 128)

    # Convert to IPv6 adddress.
    ip_obj = ipaddress.IPv6Address(int(buf, 2))
    ip_val = ipv6_norm(str(ip_obj))

    # Get result with a random host part.
    return ipv6_rand_host(ip_val, 64)

async def get_v4_lan_ips(bcast_ip, timeout=4): # pragma: no cover
    assert(type(timeout) == int)
    info_index = {
        "Windows": {
            "ping": 'ping %s -n %d' % (
                # Surrounds with double-quotes.
                win_arg_escape(bcast_ip), 
                int(timeout / 2) or 2
            ),
            "r": "(([0-9]+[.]?){4})+"
        },
        "Linux": {
            "ping": 'ping %s -b -w %d' % (
                # Surrounds with single quotes.
                nix_arg_escape(bcast_ip),
                timeout
            ),
            "r": "[(](([0-9]+[.]?){4})[)]"
        },
        "Darwin": {
            # Surrounds with double-quotes.
            "ping": 'ping %s -t %d' % (
                mac_arg_escape(bcast_ip),
                timeout
            ),
            "r": "[(](([0-9]+[.]?){4})[)]"
        },
    }

    info = info_index[platform.system()]

    # Send ping command.
    out = await cmd(info["ping"], timeout=None)

    # Get arp results.
    out = await cmd("arp -a", timeout=None)

    # Get list of results.
    results = re.findall(info["r"], out)
    if results:
        return [result[0] for result in results]
    else:
        return []

"""
- Removes %interface name after an IPv6.
- Expands shortened / or abbreviated IPs to
their longest possible form.

Why? Because comparing IPs considers IPv6s
to be 'different' if they have different interfaces
attached / missing them.

Or if you compare the same compressed IPv6 to
its uncompressed form (textually) then it
will give a false negative.
"""
def ipv6_norm(ip_val):
    ip_obj = ipaddress.ip_address(ip_val)
    if ip_obj.version == 6:
        return str(ip_obj.exploded)

    return str(ip_obj)

def ip_strip_if(ip):
    if isinstance(ip, str):
        if "%" in ip:
            parts = ip.split("%")
            return parts[0]
    
    return ip

def ip_strip_cidr(ip):
    if isinstance(ip, str):
        if "/" in ip:
            ip = ip.split("/")[0]

    return ip

def ip_norm(ip):
    # Stip interface scope id.
    ip = ip_strip_if(ip)

    # Strip CIDR.
    ip = ip_strip_cidr(ip)

    # Convert IPv6 to exploded form
    # if it's IPv6.
    ip = ipv6_norm(ip)

    return ip

def ip6_patch_bind_ip(ip_obj, bind_ip, interface):
    # Add interface descriptor if it's link local.
    if ip_obj.is_private:
        if to_s(bind_ip[0:2]).lower() == "fe":
            # Interface specified by no on windows.
            if platform.system() == "Windows":
                bind_ip = "%s%%%d" % (
                    bind_ip,
                    interface.nic_no
                )
            else:
                # Other platforms just use the name
                bind_ip = "%s%%%s" % (
                    bind_ip,
                    interface.name
                )

    return bind_ip

"""
Provides an interface that allows for bind() to be called
with its own parameters as a Route object method. Allows
the IP and port used to be accessed inside it as properties.
Otherwise defaults to using IP and port already set in class
which would only be the case if this method were used from a
Bind object and not a Route object. So a lot of hacks here.
But that's the API I wanted.
"""
def bind_closure(self):
    async def bind(port=None, ips=None):
        loop = asyncio.get_event_loop()
        port = port or self.bind_port
        if ips is not None:
            # Tested.
            nic_bind, ext_bind = self._convert_ips(
                ips=ips,
                af=self.af,
                interface=self.interface,
            )
        else:
            # Being inherited by a Route object.
            if self.nic_bind is None and self.ext_bind is None:
                nic_bind = self.nic()
                ext_bind = self.ext()
            else:
                # Use interface fields set by _convert_ips.
                # Tested.
                nic_bind = self.nic_bind
                ext_bind = self.ext_bind

        # Creates two bind tuples for nic private IP and external address.
        for bind_info in [[NIC_BIND, nic_bind], [EXT_BIND, ext_bind]]:
            bind_type, bind_ip = bind_info
            if bind_ip is None:
                # Set a blank bind tuple for this.
                log("> bind type = {} was None".format(bind_type))
                self._bind_tups[bind_type] = (None, None)
                continue

            # Add scope ID for IPv6.
            if self.af == IP6:
                # Convert bind IP to an IPAddress obj.
                ip_obj = ip_f(bind_ip)

                # Add interface descriptor if it's link local.
                is_priv = ip_obj.is_private
                not_all = bind_ip == "::" # Needed for V6 AF_ANY for some reason.
                if is_priv and not_all:
                    # Interface specified by no on windows.
                    if platform.system() == "Windows":
                        bind_ip = "%s%%%d" % (
                            bind_ip,
                            self.interface.nic_no
                        )
                    else:
                        # Other platforms just use the name
                        bind_ip = "%s%%%s" % (
                            bind_ip,
                            self.interface.name
                        )

                # Get bind info using get address.
                # This includes the special 'flow info' and 'scope id'
                
                addr_infos = await loop.getaddrinfo(
                    bind_ip,
                    port
                )

                # (Host, port, flow info, interface ID)
                if not len(addr_infos):
                    raise Exception("Can't resolve IPv6 address for bind.")
                
                self._bind_tups[bind_type] = addr_infos[0][4]

                
                if is_priv and platform.system() == "Windows":
                    self._bind_tups[bind_type] = self._bind_tups[bind_type][:3] + (self.interface.nic_no,)
            else:
                # Otherwise this is all you need.
                self._bind_tups[bind_type] = (bind_ip, port)

            # Patch for IPv6 localhost.
            if self._bind_tups[bind_type][0] == "::1":
                new_bind_tup = self._bind_tups[bind_type][:-1] + (0,)
                self._bind_tups[bind_type] = new_bind_tup

        #bind.addr = getattr(bind, 'addr', self._addr)
        self.bind_port = port
        self.nic_bind, self.ext_bind = nic_bind, ext_bind
        self.resolved = True
        return self
        
    return bind

"""
Mostly this class will not be used directly by users.
It's code is also shitty for res. Routes have superseeded this.
"""
class Bind():
    def __init__(self, interface, af, port=0, ips=None, leave_none=0):
        #if IS_DEBUG:
        #assert("Interface" in str(type(interface)))

        self.ips = ips
        self.interface = interface
        self.af = af
        self.resolved = False
        self.bind_port = port
        if leave_none:
            self.nic_bind = self.ext_bind = None
        else:
            self.nic_bind, self.ext_bind = self._convert_ips(
                ips=ips,
                af=af,
                interface=interface
            )

        # Will store a tuple that can be passed to bind.
        self._bind_tups = {NIC_BIND: None, EXT_BIND: None}
        if not hasattr(self, "bind"):
            self.bind = bind_closure(self)

    def __await__(self):
        return self.bind().__await__()

    """
    ips param can be one of any number of types. The idea is to
    extract a target IP to bind on. This is a bit excessive but
    still. Here it is. 
    """
    def _convert_ips(self, ips, af, interface=None):
        # Use interface IPs.
        ip_val = ext_bind = nic_bind = None
        if interface is not None and ips is None:
            main_route = interface.route(af)
            nic_bind = main_route.nic()
            ext_bind = main_route.ext()

            # Make it so the program crashes if they
            # try to use a default ext when fetching
            # the bind tuple.
            if af == IP6 and not interface.resolved:
                ext_bind = None
                
            return nic_bind, ext_bind 

        # Only choose the version we need.
        if ips is not None and isinstance(ips, dict):
            ip_val = ips[af]

        # Code to pass an Interface or Address
        # to initalise the ips list.
        if isinstance(ips, str):
            ip_val = ips

        # No idea what ips is.
        if ip_val is None:
            raise NotImplemented("Can't bind to that type.")

        # Convert special values in IP val.
        if ip_val in ["", "*"]:
            if af == IP6:
                nic_bind = ext_bind = ip_val = "::"
            else:
                nic_bind = ext_bind = ip_val = "0.0.0.0"
        else:
            ip_val = ip_norm(ip_val)
            ip_obj = ip_f(ip_val)
            assert(af == v_to_af(ip_obj.version))
            nic_bind = ip_val
            ext_bind = ip_val

        return nic_bind, ext_bind

    async def res(self):
        return await self.bind()

    async def start(self):
        await self.res()

    def bind_ip(self, af=None):
        af = af or self.af
        return self.nic_bind if af == IP4 else self.ext_bind

    def bind_tup(self, port=None, flag=NIC_BIND):
        # Handle loopback support.
        if flag == LOOPBACK_BIND:
            if self.af == IP6:
                return ("::1", self.bind_port)
            else:
                return ("127.0.0.1", self.bind_port)

        # Spawn a new copy of the bind tup (if needed.)
        tup = self._bind_tups[flag]
        if port is not None:
            tup = copy.deepcopy(tup)
            tup[1] = port

        # IP may not be set if invalid type of IP passed to Bind
        # and then the wrong flag type was used with it.
        if tup[0] is None:
            e = "Bind ip is none. Possibly an invalid IP "
            e += "(private and not public or visa versa) "
            e += "was passed to Bind for IPv6 causing no "
            e += "IP for the right type to be set. "
            e += "Also possible there were no link locals."
            raise Exception(e)

        #log("> binding to tup = {}".format(tup))
        return tup

    def supported(self):
        return [self.af]

async def socket_factory(route, dest_addr=None, sock_type=TCP, conf=NET_CONF):
    # Check route is bound.
    if not route.resolved:
        raise Exception("You didn't bind the route!")

    # Check addresses were processed.
    if dest_addr is not None:
        if not dest_addr.resolved:
            raise Exception("net sock factory: dest addr not resolved")
        else:
            if not dest_addr.port:
                raise Exception("net: dest port is 0!")

            # If dest_addr was a domain it will be AF_ANY.
            # We need to specify a type for the socket though.
            if route.af not in dest_addr.supported():
                raise Exception("Route af not supported by dest addr")

    # Create socket.
    sock = socket.socket(route.af, sock_type, conf["sock_proto"])

    # Useful to cleanup server sockets right away.
    if conf["linger"] is not None:
        # Enable linger and set it to its value.
        linger = struct.pack('ii', 1, conf["linger"])
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER, linger)

    # Reuse port to avoid errors.
    if conf["reuse_addr"]:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        except:
            #log("> sock fac: cant set reuse port")
            pass
            # Doesn't work on Windows.

    # Set broadcast option.
    if conf["broadcast"]:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

    # This may be set anyway by the async wrappers.
    sock.settimeout(0)

    # Bind to specific interface if set.
    # Assume Linux -- admin needed to bind on a certain interface.
    # If the interface is default for address type then no need to
    # specifically bind to it (which requires root.)
    try:
        if route.interface is not None:
            is_default = route.interface.is_default(route.af)
            if not is_default and NOT_WINDOWS:
                #log("> attemping to bind to non default iface")
                sock.setsockopt(socket.SOL_SOCKET, 25, to_b(route.interface.id))
                #log("> success: bound to non default iface")
    except Exception:
        log_exception()
        log("> couldnt bind to specific iface.")
        if sock is not None:
            sock.close()
        return None 


    # Default = use any IPv4 NIC.
    # For IPv4 -- bind address depends on destination type.
    bind_flag = NIC_BIND
    bind_tup = None
    if dest_addr is not None:
        # Get loopback working.
        if dest_addr.is_loopback:
            bind_flag = LOOPBACK_BIND

    """
        else:

            # Use global address on IPv6 global scopes.
            # Otherwise use link local or private NIC addresses.
            if dest_addr.is_public and dest_addr.chosen == IP6:
                bind_flag = EXT_BIND
    else:
        # The assumption is they want their server to be reachable.
        if route.af == IP6:
            bind_flag = EXT_BIND
    """

    try:
        sock.bind(bind_tup or route.bind_tup(flag=bind_flag))
        return sock
    except Exception:
        log(f"Could not bind to interface af = {route.af}")
        log(f"sock = {sock}")
        log(bind_tup or route.bind_tup(flag=bind_flag))
        log_exception()
        if sock is not None:
            sock.close()

async def get_high_port_socket(route, sock_type=TCP):
    # Minimal config to pass socket factory.
    conf = {
        "broadcast": False,
        "linger": None,
        "sock_proto": 0,
        "reuse_addr": True
    }

    # Get a new socket bound to a high order port.
    for i in range(0, 20):
        n = rand_rang(2000, MAX_PORT - 1000)
        await route.bind(n)
        try:
            s = await socket_factory(
                route,
                sock_type=sock_type,
                conf=conf
            )
        except:
            continue

        return s, n
    
    raise Exception("Could not bind high range port.")


from .daemon import *

class EchoServer(Daemon):
    def __init__(self):
        super().__init__()

    async def msg_cb(self, msg, client_tup, pipe):
        await async_wrap_errors(
            pipe.send(msg, client_tup)
        )

if __name__ == "__main__": # pragma: no cover
    print("See tests/test_daemon.py for code that uses this.")



"""
This is my attempt to visualize the association between private, NIC
addresses and public WAN addresses on a network interface. I have
learned the following information about network addresses:
 
    * A NIC can have one or more addresses.
    * A NIC can be assigned a block or range of addresses.
    * A NIC doesn't have to use private addresses. It's common for
    server hosts to assign the external addresses that belong
    to the server in such a way that they are used by the NIC.
    In such a case: the NICs addresses would be the same as
    how it was viewed from the external Internet.
    * A NIC can use public addresses that it doesn't own on
    the Internet. This is very bad because it means that these
    addresses will be unreachable on the Internet on that machine.
    NICs should ideally use private addresses. Or stick to IPs
    they actually can route to themselves on the Internet.
    * A NIC defines a "default" gateway to route packets to
    the Internet (which is given by network 0.0.0.0 in IPv4.)
    " The NIC can actually specify multiple default gateways.
    Each entry is a route in the route table. It will have a
    'metric' indicates its 'speed.' The route with the
    the lowest metric is chosen to route packets. TCP/IP may
    adjust the metric of routes based on network conditions.
    Thus, if there are multiple gateways for a NIC then its
    possible for the external WAN address to change under
    high network load. This is not really ideal.
 
The purpose of this module is to have easy access to the
external addresses of the machine and any associated NIC
addresses needed for Bind calls in order to use them. I
use the following simple rules to make this possible:
 
    1. All private addresses for a NIC form a group. This
    group points to the same external address for that NIC.
    2. Any public addresses are tested using STUN. If STUN
    sees the same result as the public address then the
    address is considered public and forms its own route.
    If STUN reports a different result then the address is
    being improperly used for a private NIC address. It
    thus gets added to the private group in step 1.
    3. If there is a block of public addresses to check
    only the first address is checked. If success then
    I assume the whole block is valid. Ranges of
    addresses are fully supported.
 
When it comes to complex routing tables that have
strange setups with multiple default gateways for
a NIC I am for now ignoring this possibility. I
don't consider myself an expert on networking (its
much more complex than it appears) but to directly
leverage routes in a routing table seems to me that
it would require having to work on the ethernet layer.
Something much more painful than regular sockets.

One last thing to note about routing tables: there is
a flag portion that indicates whether a route is 'up.'
If this means 'online' and 'reachable' it would be
really useful to check this to determine if a stack
supported IPv6 or IPv4 rather than trying to test it
first using STUN and waiting for a long time out.
"""

import asyncio
import copy
import ipaddress
import pprint
import inspect
from functools import total_ordering
from .ip_range import *
from .netiface_extra import *
from .upnp import *
from .address import *

"""
As there's only one STUN server in the preview release the
consensus code is not needed.
"""
ROUTE_CONSENSUS = [1, 1]

# Allows referencing a list of routes as if all WAN IPs
# were at their own index regardless of if they're in ranges.
# Will be very slow if there's a lot of hosts.
class RoutePoolIter():
    def __init__(self, rp, reverse=False):
        self.rp = rp
        self.reverse = reverse
        self.host_p = 0
        self.route_offset = 0

        # Point to the end route -- we're counting backwards.
        if self.reverse:
            self.route_offset = len(self.rp.routes) - 1

    def __iter__(self):
        return self

    def __next__(self):
        # Avoid overflow.
        if self.host_p >= self.rp.wan_hosts:
            raise StopIteration

        # Offset used for absolute position of WAN host.
        if self.reverse == False:
            host_offset = self.host_p
        else:
            host_offset = (len(self.rp) - 1) - self.host_p
        
        # Get a route object encapsulating that WAN host.
        route = self.rp.get_route_info(
            self.route_offset,
            self.host_p
        )

        # Adjust position of pointers.
        self.host_p += 1
        if self.reverse == False:
            if self.host_p >= self.rp.len_list[self.route_offset]:
                if self.route_offset < len(self.rp.routes) - 1:
                    self.route_offset += 1
        else:
            if self.host_p >= self.rp.len_list[self.route_offset]:
                if self.route_offset:
                    self.route_offset -= 1

        # Return the result.
        return route

@total_ordering
class Route(Bind):
    def __init__(self, af, nic_ips, ext_ips, interface=None, ext_check=1):
        # Sanity tests.
        assert(af in VALID_AFS)
        assert(isinstance(nic_ips, list))
        assert(isinstance(ext_ips, list))
        assert(len(ext_ips))
        assert(len(nic_ips))

        # Check value and type of ext_ip.
        assert(isinstance(ext_ips[0], IPRange))
        assert(ext_ips[0].i_ip) # IP must not be 0.
        assert(ext_ips[0].af == af)

        # Check NIC values.
        for nic_ipr in nic_ips:
            assert(isinstance(nic_ipr, IPRange))
            assert(nic_ipr.af == af)

        # Allow ext to be private if check is disabled.
        # Needed to allow for conversion from a Bind to a Route.
        if ext_check:
            assert(ext_ips[0].is_public)

        # Interface my be None.
        super().__init__(interface, af, leave_none=1)
        self.af = af
        self.nic_ips = nic_ips or []
        self.ext_ips = ext_ips or []
        self.link_locals = []

        # Maybe None for loopback interface.
        self.interface = interface
        self.route_pool = self.route_offset = self.host_offset = None

    """
    async def bind(self, port=None, ips=None):
        if self.af == IP4:
            # Patch for non-routable public IPv6s.
            print("aaa")
            print(self.ext_ips[0])
            old_ext = self.ext
            if self.ext_ips[0] not in self.nic_ips:
                print("bbb")
                for nic_ipr in self.nic_ips:
                    print(str(nic_ipr[0]))
                    if "fe80" != str(nic_ipr[0])[:4]:
                        self.ext = lambda: str(nic_ipr[0])
                        break

            self.ext = old_ext

        await bind_closure(self)(port=port, ips=ips)

    

        print(self._bind_tups)
        """

    def __await__(self):
        return self.bind().__await__()
    
    def set_link_locals(self, link_locals):
        self.link_locals = link_locals
    
    async def Address(self, dest, port):
        return await Address(dest, port, self)

    async def forward(self, port=None, proto="TCP"):
        assert(self.resolved)
        port = port or self.bind_port
        ip = self.bind_ip(self.af)
        src_addr = await Address(
            ip,
            port,
            self
        ).res()
        return await port_forward(
            interface=self.interface,
            ext_port=port,
            src_addr=src_addr,
            desc=f"P2PD {hash(src_addr.tup)}"[0:8],
            proto=proto
        )

    # TODO: document this? You probably don't want to use this.
    async def rebind(self, port=0, ips=None):
        route = copy.deepcopy(self)
        await route.bind(port=port, ips=ips)
        return route

    # A little bit nicer than accessing fields directly
    # every time just to bind to a route.
    def nic(self):
        """
        Try to select a link local (if one exists) for IPv6.
        The IPv6 proto requires at least one link local
        for core protocols like router advertisements and
        such to work properly. Assuming that IPv6 support is
        enabled on a host. If not this will raise an Exception.
        """

        """
        if self.af == IP6:
            for ipr in self.nic_ips:
                if ipr.is_private:
                    return ipr_norm(ipr)

            raise Exception("> Route.nic() with af=6 found no link-locals.")
        """

        return ipr_norm(self.nic_ips[0])

    def ext(self):
        """
        # Patch for unroutable IPv6 used as LAN IPs.
        # This is only visable to the bind() caller.
        if self.af == IP6:
            print("here")
            for stack_f in inspect.stack():
                f_name = stack_f[3]
                if f_name == "bind":
                    if self.ext_ips[0] not in self.nic_ips:
                        print("bbb")
                        for nic_ipr in self.nic_ips:
                            print("ccc")
                            print(str(nic_ipr[0]))
                            if "fe80" != ip_norm(nic_ipr[0])[:4]:
                                return ip_norm(nic_ipr[0])
        """

        return ipr_norm(self.ext_ips[0])
    
    def link_local(self):
        return ipr_norm(self.link_locals[0])

    # Test if a given IPRange is in the nic_ips list.
    def has_nic_ip(self, ipr):
        for nic_ipr in self.nic_ips:
            if nic_ipr == ipr:
                return True

        return False

    def set_offsets(self, route_offset, host_offset=None):
        self.route_offset = route_offset
        self.host_offset = host_offset

    def link_route_pool(self, route_pool):
        self.route_pool = route_pool

    def _check_extended(self):
        if self.route_pool is None:
            raise Exception("e = route_pool not linked.")

    @staticmethod
    def _convert_other(other):
        if isinstance(other, Route):
            if len(other.ext_ips):
                return other.ext_ips[0]
            else:
                return []

        if isinstance(other, IPRange):
            return other

        if isinstance(other, bytes):
            other = to_s(other)

        if isinstance(other, (str, int)):
            ipa = ipaddress.ip_address(other)
            ipr = IPRange(other, cidr=CIDR_WAN)
            return ipr

        if isinstance(other, IPA_TYPES):
            af = v_to_af(other.version)
            ipr = IPRange(other, cidr=CIDR_WAN)
            return ipr

        raise NotImplemented("Cannot convert other to IPRange in route.")

    def bad_len(self, other):
        if not len(self) or not len(other):
            return True
        else:
            return False

    # Get a list of N routes that don't use this WAN IP.
    # Incrementally adjusts route offset so its efficent.
    def alt(self, limit, exclusions=None):
        # Init storage vars.
        # Check the class has been mapped to a RoutePool.
        self._check_extended()
        routes = []
        n = 0

        # Return limit results.
        for route in self.route_pool:
            # Skip self.
            if route == self:
                continue

            # If exclude is not then get alternate route to self.
            if exclusions is not None:
                if route in exclusions:
                    continue

            # Make list of results.
            # There may be a huge number of hosts so stop at limit.
            routes.append(route)
            n += 1
            if n >= limit:
                break

        return routes

    def to_dict(self):
        nic_ips = []
        ext_ips = []
        list_infos =  [[nic_ips, self.nic_ips], [ext_ips, self.ext_ips]]
        for list_info in list_infos:
            dest_list, src_list = list_info
            for ipr in src_list:
                dest_list.append(ipr.to_dict())

        return {
            "af": int(self.af),
            "nic_ips": nic_ips,
            "ext_ips": ext_ips,
        }

    @staticmethod
    def from_dict(d):
        nic_ips = []
        ext_ips = []
        list_info =  [[nic_ips, d["nic_ips"]], [ext_ips, d["ext_ips"]]]
        for dest_list, src_list in list_info:
            for ipr_d in src_list:
                ipr = IPRange.from_dict(ipr_d)
                dest_list.append(ipr)

        af = IP4 if d["af"] == IP4 else IP6
        return Route(
            af=af,
            nic_ips=nic_ips,
            ext_ips=ext_ips
        )

    # Pickle.
    def __getstate__(self):
        return self.to_dict()

    # Unpickle.
    def __setstate__(self, state):
        o = self.from_dict(state)
        self.__dict__ = o.__dict__

    # Route != [Route, ...] = [Route, ...]
    # (max len = len(right operand))
    def __ne__(self, other):
        # Compare selfs WAN to others WAN.
        if isinstance(other, Route):
            return self.ext_ips[0] != other.ext_ips[0]

        # Otherwise get a list of routes, not matching the ones provided.
        if not isinstance(other, list):
            raise NotImplemented("Route != ? not implemented")
        else:
            return self.alt(limit=len(other), exclude_wans=other)

    # Return first route that doesn't use this same WAN IP.
    # Incrementally adjusts route offset so its efficent.
    # not Route = route_info (with different WAN to left operand.)
    def __invert__(self):
        self._check_extended()
        for route in self.route_pool:
            # If route has same external addr then skip.
            if route == self:
                continue

            return route

        return None

    def __len__(self):
        if len(self.ext_ips) == 0:
            return 0
        else:
            return len(self.ext_ips[0])

    def __repr__(self):
        return f"Route.from_dict({str(self)})"

    def __str__(self):
        return pprint.pformat(self.to_dict())

    def __eq__(self, other):
        other = Route._convert_other(other)
        if self.bad_len(other):
            return False

        return self.ext_ips[0] == other

    def __contains__(self, other):
        return self == other

    def __lt__(self, other):
        other = self._convert_other(other)
        if self.bad_len(other):
            return False

        return self.ext_ips[0] < other

    def __deepcopy__(self, memo):
        # Will fall back to the __deepcopy__ of IPRange.
        nic_ips = [copy.deepcopy(nic_ip) for nic_ip in self.nic_ips]
        ext_ips = [copy.deepcopy(ext_ips) for ext_ips in self.ext_ips]

        # Probably does nothing. YOLO.
        route = Route(self.af, nic_ips, ext_ips, self.interface)
        route.set_offsets(self.route_offset, self.host_offset)
        if self.route_pool is not None:
            route.link_route_pool(self.route_pool)
        route.ips = self.ips
        route.bind_port = self.bind_port
        route.nic_bind = self.nic_bind
        route.ext_bind = self.ext_bind
        route._bind_tups = copy.deepcopy(self._bind_tups)
        route.resolved = self.resolved
        route.set_link_locals(copy.deepcopy(self.link_locals))

        return route

class RoutePool():
    def __init__(self, routes=None, link_locals=None):
        self.routes = routes or []
        self.link_locals = link_locals or []

        # Avoid duplicates in routes.
        for route in self.routes:
            if route not in self.routes:
                self.routes.append(route)

        # Make a list of the address size for WAN portions of routes.
        # Such information will be used for dereferencing routes.
        self.len_list = []
        self.wan_hosts = 0
        for i in range(0, len(self.routes)):
            # Link route to route pool.
            self.routes[i].link_route_pool(self)
            self.routes[i].set_offsets(route_offset=i)

            # No WAN ipr section defined.
            if not len(self.routes[i].ext_ips):
                self.len_list.append(self.wan_hosts)
                continue

            # Append val to len_list = current hosts + wan hosts at route.
            next_val = self.wan_hosts + len(self.routes[i].ext_ips[0])
            self.len_list.append(next_val)
            self.wan_hosts = next_val

        # Index into the routes list.
        self.route_index = 0

        # Simulate 'removing' past elements.
        self.pop_pointer = 0

    def to_dict(self):
        routes = []
        for route in self.routes:
            routes.append(route.to_dict())

        return routes

    @staticmethod
    def from_dict(route_dicts):
        routes = []
        for route_dict in route_dicts:
            routes.append(Route.from_dict(route_dict))

        return RoutePool(routes)

    # Pickle.
    def __getstate__(self):
        return self.to_dict()

    # Unpickle.
    def __setstate__(self, state):
        o = self.from_dict(state)
        self.__dict__ = o.__dict__

    def locate(self, other):
        for route in self.routes:
            if route == other:
                return route

        return None

    # Is a route in this route pool?
    def __contains__(self, other):
        route = self.locate(other)
        if route is not None:
            return True
        else:
            return False

    # Simulate fetching a route off a stack of routes.
    # Just hides certain pointer offsets when indexing, lel.
    def pop(self):
        if self.pop_pointer >= self.wan_hosts:
            raise Exception("No more routes.")

        ret = self[self.pop_pointer]
        self.pop_pointer += 1

        return ret

    def get_route_info(self, route_offset, abs_host_offset):
        # Route to use for the WAN addresses.
        assert(route_offset <= (len(self.routes) - 1))
        route = self.routes[route_offset]
        
        # Convert host_offset to a index inside route's WAN subnet.
        prev_len = self.len_list[route_offset - 1] if route_offset else 0
        rel_host_offset = abs_host_offset - prev_len
        assert(rel_host_offset <= self.len_list[route_offset] - 1)

        # Get references to member objs.
        wan_ipr = route.ext_ips[0]
        nic_ipr = route.nic_ips[0]

        # For pub ranges assigned to NIC -- they will line up.
        # For N or more private addressess -> a WAN = probably won't.
        # In such a case it doesn't matter as any NIC IP = the same WAN.
        assert(rel_host_offset + 1 <= self.wan_hosts)
        assert(len(wan_ipr))
        assert(len(nic_ipr))
        rel_host_offset = rel_host_offset % self.wan_hosts
        
        # Build a route corrosponding to these offsets.
        wan_ip = IPRange(str(wan_ipr[rel_host_offset]), cidr=CIDR_WAN)
        new_route = Route(
            af=route.af,
            nic_ips=copy.deepcopy(route.nic_ips),
            ext_ips=[wan_ip],
            interface=route.interface
        )
        new_route.set_offsets(route_offset, abs_host_offset)
        new_route.link_route_pool(self)

        return new_route

    def __len__(self):
        return self.wan_hosts

    def __getitem__(self, key):
        # Possible due to pop decreasing host no.
        if not self.wan_hosts:
            return []

        if isinstance(key, slice):
            start, stop, step = key.indices(len(self))
            return [self[i] for i in range(start, stop, step)]
        elif isinstance(key, int):
            # Convert negative index to positive.
            # Sorted_search doesn't work with negative indexex.
            if key < 0:
                key = key % self.wan_hosts

            route_offset = sorted_search(self.len_list, key + 1)
            if route_offset is None:
                return None
            else:
                return self.get_route_info(route_offset, key)
        elif isinstance(key, tuple):
            return [self[x] for x in key]
        else:
            raise TypeError('Invalid argument type: {}'.format(type(key)))

    def __iter__(self):
        return RoutePoolIter(self)

    def __reversed__(self):
        return RoutePoolIter(self, reverse=True)

def rp_from_fixed(fixed, interface, af): # pragma: no cover
    """
    [
        [
            nics [[ip, opt netmask], ...],
            exts [[ip]]
        ],
        route ...
    ]
    """

    routes = []
    for route in fixed:
        nic_iprs = []
        ext_iprs = []
        for meta in [[nic_iprs, route[0]], [ext_iprs, route[1]]]:
            dest, nic_infos = meta
            for nic_info in nic_infos:
                ip = nic_info[0]
                netmask = None
                if len(nic_info) == 2:
                    netmask = nic_info[1]

                ipr = IPRange(ip, netmask=netmask)
                dest.append(ipr)

        route = Route(af, nic_iprs, ext_iprs, interface)
        routes.append(route)

    return RoutePool(routes)

"""
When it comes to IPs assigned to a NIC its possible
to assign 'public' IPs to it directly. You often
see this setup on servers. In this case you know
that not only can you use the public addresses
directly in bind() calls -- but you know that
the server's corrosponding external IP will be
what was used in the bind() call. Very useful.

The trouble is that network interfaces happily
accept 'external IPs' or IPs outside of the
typical 'private IP' range for use on a NIC or
LAN network. Obviously this is a very bad idea
but in the software it has the result of
potentially assuming that an IP would end up
resulting in a particular external IP being used.

The situation is not desirable when building
a picture of a network's basic routing makeup.
I've thought about the problem and I don't see
a way to solve it other than to measure how a
route's external address is perceived from the
outside world. Such a solution is not ideal but
at least it only has to be done once.
"""
async def ipr_is_public(nic_ipr, stun_client, route_infos, stun_conf, is_default=False):
    # Use this IP for bind call.
    src_ip = ip_norm(str(nic_ipr[0]))
    log(F"src_ip = {src_ip}")
    local_addr = await Bind(
        stun_client.interface,
        af=stun_client.af,
        port=0,
        ips=src_ip
    ).res()

    # Get external IP and compare to bind IP.
    wan_ip = await stun_client.get_wan_ip(
        local_addr=local_addr,
        conf=stun_conf
    )

    # Record this wan_ip.
    ext_ipr = IPRange(wan_ip, cidr=CIDR_WAN)
    if ext_ipr not in route_infos:
        route_infos[ext_ipr] = []

    # Determine if public address is assigned to interface.
    if src_ip != wan_ip:
        # In a 'public range' but used privately.
        # Yes, this is silly but possible.
        nic_ipr.is_private = True
        nic_ipr.is_public = False
        if "default" not in route_infos:
            route_infos["default"] = ext_ipr
    else:
        # Otherwise routable, public IP used as a NIC address.
        # ext_ipr == nic_ipr.
        nic_ipr.is_private = False
        nic_ipr.is_public = True

    # Save details for this route.
    route_infos[ext_ipr].append(nic_ipr)

async def get_routes(interface, af, skip_resolve=False, skip_bind_test=False, netifaces=None, stun_client=None):
    from .stun_client import STUNClient, STUN_CONF

    # Settings to use for external addresses over STUN.
    if interface is not None:
        log(f">get routes for {interface.name} {af}")
        nic_id = interface.name
    else:
        nic_id = None

    # Route-specific config options for STUN.
    stun_conf = copy.deepcopy(STUN_CONF)
    stun_conf["retry_no"] = 2 # Slower but more fault-tolerant.
    stun_conf["consensus"] = ROUTE_CONSENSUS # N of M for a result.
    loop = asyncio.get_event_loop()
    link_locals = []
    route_infos = {}
    nic_iprs = []
    routes = []

    # Other important variables.
    stun_client = stun_client or STUNClient(interface, af=af)
    netifaces_af = af_to_netiface(af)
    if_addresses = netifaces.ifaddresses(nic_id)
    if netifaces_af in if_addresses:
        bound_addresses = if_addresses[netifaces_af]
        tasks = []
        first_private = True
        for info in bound_addresses:
            nic_ipr = await netiface_addr_to_ipr(af, info, interface, loop, skip_bind_test)
            log(f"Nic ipr in route = {nic_ipr}")
            if nic_ipr is None:
                continue

            # Include link locals in their own set.
            if ip_norm(nic_ipr[0])[:4] == "fe80":
                link_locals.append(nic_ipr)
                continue

            # All private IPs go to the same route.
            # The interface has one external WAN IP.
            log(f"Nic ipr is private = {nic_ipr.is_private}")
            if nic_ipr.is_private:
                # Ensure the external address is fetched at least once
                # assuming there's only private addresses.
                # All private NICs will point to this.
                if first_private:
                    tasks.append(
                        async_wrap_errors(
                            ipr_is_public(nic_ipr, stun_client, route_infos, stun_conf),
                            timeout=15
                        )
                    )

                    first_private = False
                else:
                    nic_iprs.append(nic_ipr)
            else:
                if skip_resolve == False:
                    # Determine if this address is really public.
                    tasks.append(
                        async_wrap_errors(
                            ipr_is_public(nic_ipr, stun_client, route_infos, stun_conf),
                            timeout=15
                        )
                    )
                else:
                    # Assume IP is public and routable.
                    # Useful for manually configured interfaces.
                    nic_ipr.is_private = False
                    nic_ipr.is_public = True
                    route = Route(
                        af=af,
                        nic_ips=[nic_ipr],
                        ext_ips=[copy.deepcopy(nic_ipr)],
                        interface=stun_client.interface
                    )

                    routes.append(route)

        # Process pending tasks.
        if len(tasks):
            await asyncio.gather(*tasks)

            # Choose a random ext address for the default.
            if "default" not in route_infos:
                route_infos["default"] = list(route_infos.keys())[0]

            # Add private NIC iprs to default route.
            default_route = route_infos["default"]
            default_nics = route_infos[default_route]
            for nic_ipr in nic_iprs:
                default_nics.append(nic_ipr)

            # Default route will be added first.
            routes.append(Route(af, default_nics, [default_route], interface))
            del route_infos["default"]

            # Convert route infos to routes and save them.
            for ext_ipr in route_infos:
                if ext_ipr != default_route:
                    routes.append(
                        Route(
                            af,
                            route_infos[ext_ipr],
                            [ext_ipr],
                            interface
                        )
                    )

        # Add link locals to routes.
        [r.set_link_locals(link_locals) for r in routes]
        
    return [routes, link_locals]

async def Routes(interface_list, af, netifaces, skip_resolve=False):
    # Optimization: check if an AF has a default gateway first.
    # If it doesn't return an empty route pool for AF.
    if not is_af_routable(af, netifaces):
        log("> af {} has no default route".format(af))
        return RoutePool([])

    # Hold results.
    results = []
    tasks = []

    # Copy route pool from Interface if it already exists.
    # Otherwise schedule task to get list of routes.
    for iface in interface_list:
        tasks.append(
            get_routes(iface, af, skip_resolve, netifaces=netifaces)
        )

    # Tasks that need to be run.
    # Cmbine with results -- if any.
    link_locals = []
    if len(tasks):
        ret_lists = await asyncio.gather(*tasks)
        for ret_list in ret_lists:
            if len(ret_list) != 2:
                continue

            results = results + ret_list[0]
            link_locals = link_locals + ret_list[1]

    # Wrap all routes in a RoutePool and return the result.
    return RoutePool(results, link_locals)

# Combine all routes from interface into RoutePool.
def interfaces_to_rp(interface_list):
    rp = {}
    for af in VALID_AFS:
        route_lists = []
        for iface in interface_list:
            if af not in iface.rp:
                continue

            route_lists.append(
                copy.deepcopy(iface.rp[af].routes)
            )

        routes = sum(route_lists, [])
        rp[af] = RoutePool(routes)

    return rp

# Converts a Bind object to a Route.
# Interface for bind object may be None if it's loopback.
async def bind_to_route(bind_obj):
    if not isinstance(bind_obj, Bind):
        raise Exception("Invalid obj type passed to bind_to_route.")

    """
    nic_bind = 1 -- ipv4 nic or ipv6 link local
    ext_bind = 2 -- ipv4 external wan ip / ipv6 global ip
        black hole ip if called with no ips start_local

    ips = both set to ips value
    nic_bind or ext_bind based on dest address in sock_factory
    """
    assert(bind_obj.resolved)
    interface = bind_obj.interface
    nic_bind = bind_obj._bind_tups[NIC_BIND][0]
    ext_bind = bind_obj._bind_tups[EXT_BIND][0]
    af = bind_obj.af
    assert(interface.resolved)

    """
    If the ext_bind contains a valid public address then
    use this directly for the ext_ipr in the Route obj.
    Otherwise attempt to find a pre-existing route in
    the Interface route pool that has the same nic_bind
    and use it's associated ext_ipr.
    """
    ext_set = 0
    nic_ipr = IPRange(nic_bind)
    ext_ipr = IPRange(ext_bind)
    if not ext_ipr.is_public:
        if interface is not None:
            # Check all routes for a matching NIC IPR.
            for hey_route in interface.rp[af].routes:
                # Check all NIC IPRs.
                for nic_hey in hey_route.nic_ips:
                    # NIC IPR found in the route entries.
                    # Use the routes EXT.
                    if nic_ipr in nic_hey:
                        ext_ipr = hey_route.ext_ips[0]
                        ext_ipr = copy.deepcopy(ext_ipr)
                        ext_set = 1
                        break

                if ext_set:
                    break

    # Build route object.
    route = Route(
        af=af,
        nic_ips=[nic_ipr],
        ext_ips=[ext_ipr],
        interface=interface,
        ext_check=0
    )

    # Bind to port in route.
    await route.bind(port=bind_obj.bind_port)
    return route

if __name__ == "__main__": # pragma: no cover
    from .interface import Interface

    async def test_get_routes(): # pragma: no cover
        internode_iface = Interface("enp3s0")
        starlink_iface = Interface("wlp2s0")
        iface_list = [internode_iface, starlink_iface]
        """
        af = IP4
        rp = await Routes(iface_list, af)
        r1 = rp.routes[0]
        nr1, _ = ~r1

        # Should compare two routes WAN portions.
        assert(r1 != nr1)
        assert(r1 == r1)

        r_list = r1 != [r1]
        assert(r_list[0][0] != r1)
        """

        # Test no WAN route.
        af = IP6
        rp = await Routes(iface_list, af)
        r1 = rp.routes[0]

        # When resolving a route that isnt supported = slow
        # any way to get it to return faster?

        return

        ra, rb = rp.routes
        for r in rp.routes:
            print(r)
            print(r.nic_ips)
            print(r.ext_ips)

        ref_route_a = rp[0]
        print(rp.routes)

        print(ref_route_a)
        return

        ipr = IPRange("192.168.0.0", "255.255.255.0")
        r = RoutePool([ipr])
        #ipr2 = copy.deepcopy(ipr)

        print(id(ipr.ip))
        print(id(r.routes[0].ip))

        return
        routes = await get_routes(iface, IP4)
        print(routes)
        route = routes[0]
        print(route.nic_ips)
        print(route.ext_ips)

    async_test(test_get_routes)





import random
import binascii
from p2pd import *

async def example():
    # Open default interface.
    # Get a route for the first AF supported.
    i = await Interface().start()
    route = await i.route().bind()
    #
    # Open a UDP pipe to stunprotocol.org.
    # Subscribe to all messages.
    pipe = await pipe_open(
        UDP,
        route,
        await Address("stunserver.stunprotocol.org", 3478, route).res()
    )
    #
    # Build a STUN request and send it.
    msg_id = ''.join([str(random.randrange(10, 99)) for _ in range(16)])
    req_hex = "00010000" + msg_id
    req_buf = binascii.unhexlify(req_hex)
    await pipe.send(req_buf)
    #
    # Get the response.
    resp = await pipe.recv()
    await pipe.close()

if __name__ == '__main__':
    async_test(example)


from .address import *
from .interface import *
from .base_stream import *

DAEMON_CONF = dict_child({
    "reuse_addr": True
}, NET_CONF)

"""
interface = listen on all addresses given by af
bind = may take specific addresses to listen to -- listen only on af addresses
targets = [interface, bind, ...]

protocols to use for the server.
protos = [TCP, UDP]

list of ports to listen on. may contain ranges.
ports = [n, [x, y], ...]

called on new messages from a stream.
process the message.
f = lambda msg, stream: ...
"""
def get_listen_port(server_info):
    return server_info[2].sock.getsockname()[1]

class Daemon():
    def __init__(self, interfaces=None, conf=DAEMON_CONF):
        self.conf = conf
        self.servers = []
        self.restricted_proto = None
        self.interfaces = interfaces or []
        self.iface_lookup = {}
        for iface in self.interfaces:
            self.iface_lookup[iface.name] = iface
        self.rp = None

    def get_listen_port(self):
        return get_listen_port(self.servers[0])

    def set_rp(self, rp):
        self.rp = rp

    def restrict_proto(self, proto):
        self.restricted_proto = proto

    """
    Overwritten by children.
    msg = data received from server
    client_tup = (remote ip, remote port) of client socket.
    pipe = BaseProto object
    """
    def msg_cb(self, msg, client_tup, pipe):
        # Overwritten by inherited classes.
        print("This is a default proto msg_cb! Specify your own in a child class.")
        pass

    def up_cb(self, msg, client_tup, pipe):
        pass
    
    async def _listen(self, target, port, proto, msg_cb=None, up_cb=None):
        msg_cb = msg_cb or self.msg_cb
        up_cb = up_cb or self.up_cb

        # Protocol not supported.
        if self.restricted_proto is not None:
            if self.restricted_proto != proto:
                raise Exception("That protocol is not supported for this server.")
    
        # Convert Bind to routes.
        routes = []

        # Target is already a route - record.
        if isinstance(target, Route):
            routes.append(target)
        else:
            # Route inherits Bind so this would also be True for
            # a Route object. Check Route first to avoid this.
            if isinstance(target, Bind):
                if target.af == AF_ANY:
                    af_list = VALID_AFS
                else:
                    af_list = [target.af]

                # Make a route for each AF.
                for af_val in af_list:
                    # Interface does not support AF.
                    # None Interface if using loopback.
                    if target.interface is not None:
                        if af_val not in target.interface.what_afs():
                            log("> bind inst: af_val not in iface afs")
                            continue


                        # Gets the route associated with the bind IP.
                        route = await bind_to_route(target)

                        # Record route.
                        routes.append(route)


        # Start server on every address.
        assert(len(routes))
        for route in routes:
            if not route.resolved:
                await route.bind(port)

            log('starting server {}:{} p={}, af={}'.format(
                route.bind_ip(),
                route.bind_port,
                proto,
                route.af
            ))

            # Link route to a route pool.
            # So it can do cool tricks.
            if self.rp is not None:
                route.link_route_pool(self.rp[route.af])

            # Save list of servers.
            listen_task = pipe_open(proto, route=route, msg_cb=msg_cb, up_cb=up_cb, conf=self.conf)
            base_proto = await listen_task
            if base_proto is None:
                raise Exception("Could not start server.")

            self.servers.append([ route, proto, base_proto, listen_task ])
    
    # [[Bound, proto], ...]
    # Allows for servers to be bound to specific addresses and transports.
    # The other start method is more general.
    async def listen_specific(self, targets, msg_cb=None, up_cb=None):
        up_cb = up_cb or self.up_cb
        msg_cb = msg_cb or self.msg_cb
        targets = [targets] if not isinstance(targets, list) else targets
        for listen_info in targets:
            # Unpack.
            bound, proto = listen_info

            # Start server.
            await self._listen(
                target=bound,
                proto=proto,
                port=bound.bind_port,
                msg_cb=msg_cb,
                up_cb=up_cb
            )

    # Start all the servers listening.
    # All targets are started on the same list of ports and protocols.
    # A more general version of the above function.
    async def listen_all(self, targets, ports, protos, af=AF_ANY, msg_cb=None, up_cb=None, error_on_af=0):
        up_cb = up_cb or self.up_cb
        msg_cb = msg_cb or self.msg_cb
        targets = [targets] if not isinstance(targets, list) else targets
        routes = []
        for target in targets:
            # Convert Interface to route.
            if isinstance(target, Interface):
                # List of AFs to use.
                if af == AF_ANY:
                    af_list = target.what_afs()
                else:
                    af_list = [af]

                # If Interface doesn't support AF
                # do we throw an error or skip it?
                for af_val in af_list:
                    if af_val not in target.what_afs():
                        log("> listen all af not in what afs")
                        if error_on_af:
                            e = "IF {} doesn't support AF {}".format(
                                target.id,
                                af
                            )
                            raise Exception(e)

                # Build routes from afs + interface.
                for af_val in af_list:
                    if not len(target.rp[af_val].routes):
                        continue

                    # First route in pool for that AF.
                    # Makes a copy of the route.
                    route = target.route(af_val)
                    routes.append(route)

            # Convert RoutePool to Routes.
            # Makes new instances of the routes.
            if isinstance(target, RoutePool):
                # There could be massive blocks of IPs in a RoutePool.
                # It's not practical to index all of them.
                routes += target[:10]

            # Route.
            if isinstance(target, Route):
                routes.append(
                    copy.deepcopy(target)
                )

        # Targets are Interfaces to listen on or Bind objects.
        for route in routes:
            # Loop over all the listen ports.
            for port_info in ports:
                # Ports may be a single port or a range.
                if type(port_info) == list:
                    port_start, port_end = port_info
                else:
                    port_end = port_start = port_info

                # Treat ports as a range.
                for port in range(port_start, port_end + 1):
                    # Start the server on bind address, port, and proto.
                    for proto in protos:
                        route = copy.deepcopy(route)
                        await route.bind(port=port, ips=route.ips)
                        await self._listen(route, port, proto, msg_cb)

                        # Bind to additional link-local for IPv6.
                        if route.af == IP6:
                            try:
                                route = copy.deepcopy(route)
                                route = await route.bind(port=port, ips=route.link_local())
                                await self._listen(target=route, port=port, proto=proto, msg_cb=msg_cb, up_cb=up_cb)
                            except Exception:
                                log_exception()
                                pass
                                # May already be started -- ignore.

        return self

    async def close(self):
        for server_info in self.servers:
            _, _, server, _ = server_info
            await server.close()

        self.servers = []

if __name__ == "__main__": # pragma: no cover
    pass
    #async_test(test_tcp_punch)


import asyncio
from .p2p_pipe import *

def parse_mappings(self, parts):
    if len(parts) != 9:
        raise Exception("Invalid length for mappings msg.")

    # Extract fields from message.
    r = parse_punch_response(parts)
    p2p_dest = work_behind_same_router(self.p2p_addr, r["src_addr"])

    # Check the address family is valid.
    if r["af"] not in VALID_AFS:
        raise Exception("Invalid af for mappings msg.")

    # Check their chosen interface offset for ourselves is valid.
    if r["if"]["us"] > (len(self.if_list) - 1):
        raise Exception("Invalid if us offset for mappings.")

    # Check their used interface offset is valid.
    their_if_infos = r["src_addr"][r["af"]]
    if r["if"]["them"] > (len(their_if_infos) - 1):
        raise Exception("Invalid if them offset for mappings.")

    # Return main fields.
    their_if_info = their_if_infos[r["if"]["them"]]
    return r, p2p_dest, their_if_infos, their_if_info

async def signal_protocol(self, msg, signal_pipe):
    # Convert to string because this is a plain-text protocol.
    if isinstance(msg, memoryview):
        msg = to_s(msg.tobytes())
    else:
        msg = to_s(msg)

    # Split msg into parts.
    log(f"> signal proto msg = {msg}")
    parts = msg.split(" ")
    cmd = parts[0]

    # Basic echo protocol for testing.
    if cmd == "ECHO":
        if len(parts) >= 2:
            if isinstance(parts[1], memoryview):
                chan_dest = parts[1].tobytes()
            else:
                chan_dest = parts[1]


            # cmd sp node_id sp msg
            offset = (6 + len(chan_dest))
            out = msg[offset:]
            if len(out):
                if isinstance(out, memoryview):
                    out = out.tobytes()

                await signal_pipe.send_msg(out, chan_dest)

        return

    # Reverse connect signal.
    if cmd == "P2P_DIRECT":
        if len(parts) != 4:
            log("> invalid p2p direct msg recv.")
            return 1

        # Process message fields.
        pipe_id, proto, addr_bytes = parts[1], parts[2], parts[3]
        proto = PROTO_LOOKUP[proto]
        pipe_id = to_b(pipe_id)
        addr_bytes = to_b(addr_bytes)
        p2p_dest = parse_peer_addr(addr_bytes)
        p2p_dest = work_behind_same_router(self.p2p_addr, p2p_dest)
        log(f"p2p direct proto no = {proto}")

        # Connect to chosen address.
        p2p_pipe = P2PPipe(self)
        try:
            pipe = await asyncio.wait_for(
                p2p_pipe.direct_connect(p2p_dest, pipe_id, proto=proto),
                10
            )
        except Exception as e:
            log("p2p direct timeout in node.")
            log_exception()
            return
        
        # Setup pipe reference.
        if pipe is not None:
            log("p2p direct in node got a valid pipe.")

            # Record pipe reference.
            self.pipes[pipe_id] = pipe

            # Add cleanup callback.
            pipe.add_end_cb(self.rm_pipe_id(pipe_id))

        return

    # Request to start TCP hole punching.
    if cmd == "INITIAL_MAPPINGS":
        # Parse mappings to dict.
        ret = parse_mappings(self, parts)
        r, p2p_dest, their_if_infos, their_if_info = ret

        # Create hole punching client.
        interface = self.if_list[r["if"]["us"]]
        stun_client = STUNClient(interface=interface, af=r["af"])
        recipient = self.tcp_punch_clients[r["if"]["us"]]

        # Calculate punch mode.
        their_addr = await Address(
            str(their_if_info["ext"]),
            80,
            interface.route(r["af"])
            ).res()
        punch_mode = recipient.get_punch_mode(their_addr)
        if punch_mode == TCP_PUNCH_REMOTE:
            use_addr = str(their_if_info["ext"])
        else:
            use_addr = str(their_if_info["nic"])
        
        # Step 2 -- exchange initiator mappings with recipient.
        punch_ret = await recipient.proto_recv_initial_mappings(
            use_addr,
            their_if_info["nat"],
            r["src_addr"]["node_id"],
            r["pipe_id"],
            r["predictions"],
            stun_client,
            r["ntp_time"],
            mode=punch_mode
        )

        # Build second (optional) punch message for peer.
        out = build_punch_response(
            b"UPDATED_MAPPINGS",
            r["pipe_id"],
            punch_ret,
            self.addr_bytes,
            r["af"],
            r["if"]["us"], # Which iface we're using from our addr.
            r["if"]["them"] # Which iface they should use.
        )

        # Send first protocol signal message to peer.
        send_task = asyncio.create_task(
            async_wrap_errors(
                signal_pipe.send_msg(
                    out,
                    to_s(r["src_addr"]["node_id"])
                )
            )
        )

        # Do the hole punching.
        try:
            pipe = await asyncio.wait_for(
                get_tcp_hole(
                    PUNCH_RECIPIENT,
                    r["pipe_id"],
                    r["src_addr"]["node_id"],
                    recipient,
                    self
                ),
                30
            )
        except asyncio.TimeoutError:
            log("node tcp punch timeout.")
            return

        return

    # Additional info for doing TCP hole punching.
    if cmd == "UPDATED_MAPPINGS":
        # Unpack mapping fields and parse.
        ret = parse_mappings(self, parts)
        r, p2p_dest, their_if_infos, their_if_info = ret

        # Make a STUN client that can get mappings.
        # This actually shouldn't be needed.
        dest_s = str(their_if_info["ext"])
        af = af_from_ip_s(dest_s)
        interface = self.if_list[r["if"]["us"]]
        stun_client = STUNClient(interface, af)

        # Update received mappings.
        # This is an optional step that can improve connect success.
        initiator = self.tcp_punch_clients[r["if"]["us"]]
        ret = await initiator.proto_update_recipient_mappings(
            r["src_addr"]["node_id"],
            r["pipe_id"],
            r["predictions"],
            stun_client
        )

        return

    """
    Requests that a peer use a specified TURN server to connect
    back to a source peer. The peer provides it's 'mapped address'
    -- the external address of the peer seen from the TURN server's
    perspective. They are expected to 'white list' this address.
    A 'relay address' is also specified for sending messages back
    to the source. Towards the end this node will exchange its own
    mapped and relay address back to the source.
    """
    if cmd == "TURN_REQUEST":
        if len(parts) != 12:
            log("> turn_req: invalid parts len")
            return
        
        # Extract all fields from the signal msg.
        pipe_id = to_b(parts[1])
        af = int(parts[2])
        their_if_index = int(parts[3])
        our_if_index = int(parts[4])
        src_addr_bytes = to_b(parts[5])
        peer_ip = parts[6]
        peer_port = int(parts[7])
        relay_ip = parts[8]
        relay_port = int(parts[9])
        turn_server_index = int(parts[10])
        turn_client_index = int(parts[11])

        # Check turn server index.
        if not in_range(turn_server_index, [0, len(TURN_SERVERS) - 1]):
            log(f"> turn req: servers offset {turn_server_index}")
            return
        else:
            turn_server = TURN_SERVERS[turn_server_index]

        # Check address family is valid.
        if af not in VALID_AFS or af not in turn_server["afs"]:
            log("> turn_req: invalid af")
            return
        
        # Check interface index is valid.
        if not in_range(our_if_index, [0, len(self.if_list) - 1]):
            log("> turn_req: invalid if_index")
            return

        # Check ports are valid.
        for port in [relay_port]:
            if not in_range(port, [1, MAX_PORT]):
                log("> turn_req: invalid port")
                return

        # See if TURN server is already connected.
        interface = self.if_list[our_if_index]
        turn_client = self.find_turn_client(turn_server, interface=interface)
        if turn_client is None:
            # Resolve the TURN address.
            route = await interface.route(af).bind()
            turn_addr = await Address(
                turn_server["host"],
                turn_server["port"],
                route
            ).res()

            # Make a TURN client instance to whitelist them.
            turn_client = TURNClient(
                route=route,
                turn_addr=turn_addr,
                turn_user=turn_server["user"],
                turn_pw=turn_server["pass"],
                turn_realm=turn_server["realm"],
                msg_cb=self.msg_cb
            )

            # Start the TURN client.
            try:
                await asyncio.wait_for(
                    turn_client.start(),
                    10
                )
            except asyncio.TimeoutError:
                log("Turn client start timeout in node.")
                return

            # Set new TURN client.
            self.turn_clients.append(turn_client)

        # Resolve the peer address.
        # The address here is their XorMappedAddress.
        # The external address of the peer from the TURN server's perspective.
        route = interface.route(af)
        peer_addr = await Address(
            str(peer_ip),
            peer_port,
            route
        ).res()

        # Resolve relay address.
        relay_addr = await Address(
            relay_ip,
            relay_port,
            route
        ).res()

        # White list peer.
        try:
            await asyncio.wait_for(
                turn_client.accept_peer(peer_addr.tup, relay_addr.tup),
                6
            )
        except asyncio.TimeoutError:
            log("node turn accept peer timeout.")
            return

        # Record the pipe internally.
        client_tup = await turn_client.client_tup_future
        our_relay_tup = await turn_client.relay_tup_future
        self.pipes[pipe_id] = turn_client
        log("> turn_req: our relay tup = {}:{}".format(
            *our_relay_tup
        ))

        # Form response with our addressing info.
        out = b"TURN_RESPONSE %s %s %d %d %s %d %s %d %d" % (
            pipe_id,
            self.node_id,
            af,
            their_if_index,

            # Our own relay addr to route messages to us.
            to_b(our_relay_tup[0]),
            our_relay_tup[1],

            # Our XorMappedAddress.
            to_b(client_tup[0]),
            client_tup[1],

            # Their client to use.
            turn_client_index
        )
        
        # Send response to recipient.
        p2p_src_addr = parse_peer_addr(src_addr_bytes)
        await signal_pipe.send_msg(
            out,
            to_s(p2p_src_addr["node_id"])
        )

    """
    The peer that you requested to contact you back via TURN
    has sent you back this response. The response includes their
    mapped address and their relay address. With this info
    both peers can now start sending messages via each others
    relay addresses and the correct permissions are in place to
    let the packets through. The peers will receive replies from
    the TURN server on the TURN server's regular port. The replies
    will be Send indications with a data attribute and a
    XorPeerAddress attribute that specifies the peer address tuple
    of the packet sender -- which we discard if it doesn't match.
    """
    if cmd == "TURN_RESPONSE":
        # Invalid packet.
        if len(parts) != 10:
            log("> turn_res: invalid parts len")
            return

        # Name all the parts and type convert.
        pipe_id = to_b(parts[1])
        node_id = to_b(parts[2])
        af = int(parts[3])
        if_index = int(parts[4])
        relay_ip = parts[5]
        relay_port = int(parts[6])
        client_ip = IPRange(parts[7])
        client_port = int(parts[8])
        turn_client_index = int(parts[9])

        # Check pipe_id exists.
        if pipe_id not in self.pipe_events:
            log("> turn_res: pipe id not in events")
            return
        if pipe_id not in self.expected_addrs:
            log("> turn_res: pipe id not in turn pending")
            return

        # Check the IP matches what we expect.
        found_exts = self.expected_addrs[pipe_id]
        if client_ip not in found_exts:
            log("> turn_res: client_ip != found ext")
            return

        # Validate ports.
        for port in [relay_port, client_port]:
            if not in_range(port, [1, MAX_PORT]):
                log("> turn_res: invalid port")
                return

        # Validate address family.
        if af not in VALID_AFS:
            log("> turn_res: invalid af")
            return

        # Invalid if index.
        if not in_range(if_index, [0, len(self.if_list) - 1]):
            log("> turn_res: if index invalid.")
            return

        # Get turn client.
        if not in_range(turn_client_index, [0, len(self.turn_clients) - 1]):
            log("> turn_resp: invalid turn clietn index.")
            return

        # Get turn client reference.
        turn_client = self.turn_clients[turn_client_index]
        if turn_client is None:
            log("> turn_res: turn client none")
            return

        # Notify waiters that we received the relay address.
        client_tup = (str(client_ip), client_port)
        relay_tup = (relay_ip, relay_port)
        await turn_client.accept_peer(client_tup, relay_tup)
        turn_client.node_events[to_s(node_id)].set()
        return

async def node_protocol(self, msg, client_tup, pipe):
    log(f"> node proto = {msg}, {client_tup}")

    # Execute any custom msg handlers on the msg.
    run_handlers(pipe, self.msg_cbs, client_tup, msg)

    # Execute basic services of the node protocol.
    parts = msg.split(b" ")
    cmd = parts[0]

    # Basic echo server used for testing networking.
    if cmd == b"ECHO":
        if len(msg) > 5:
            await pipe.send(memoryview(msg)[5:], client_tup)

        return

    # This connection was in regards to a request.
    if cmd == b"ID":
        # Invalid format.
        if len(parts) != 2:
            log("ID: Invalid parts len.")
            return 1

        # If no ones expecting this connection its a reverse connect.
        pipe_id = parts[1]
        pipe.add_end_cb(self.rm_pipe_id(pipe_id))
        if pipe_id not in self.pipe_events:
            assert(isinstance(pipe_id, bytes))
            log(f"pipe = '{pipe_id}' not in pipe events. saving.")
            self.pipes[pipe_id] = pipe
        else:
            # Is this IP expected?
            if pipe_id not in self.expected_addrs:
                log("ID: pipe_id not in expected_addrs.")
                return 2

            # Check remote address is right.
            exts = self.expected_addrs[pipe_id]
            ipr = IPRange(client_tup[0])
            if ipr not in exts:
                log("ID: ipr not in expected addrs.")
                return 3

            # Pipe already saved.
            pipe_event = self.pipe_events[pipe_id]
            if pipe_event.is_set():
                log("ID: pipe event not set.")
                return 4

            # Save pipe and notify any waiters about it.
            self.pipes[pipe_id] = pipe
            pipe_event.set()


from p2pd.test_init import *
from p2pd import IPRange, Bind, Route, Interface
from p2pd.net import BLACK_HOLE_IPS, ip_norm, DUEL_STACK, IP6, IP4
from p2pd.net import NIC_BIND, EXT_BIND, TCP, socket_factory

class TestBind(unittest.IsolatedAsyncioTestCase):
    async def test_bind_closure(self):
        pass

        i = await Interface().start_local()
        b = Bind(i, None, leave_none=1)

        b.af = IP4
        out = (await bind_closure(b)(80, "127.0.0.1"))._bind_tups
        assert(out == {1: ('127.0.0.1', 80), 2: ('127.0.0.1', 80)})

        out = (await bind_closure(b)(80, "192.168.0.1"))._bind_tups
        assert(out == {1: ('192.168.0.1', 80), 2: ('192.168.0.1', 80)})

        out = (await bind_closure(b)(80, "8.8.8.8"))._bind_tups
        assert(out == {1: ('8.8.8.8', 80), 2: ('8.8.8.8', 80)})

        out = (await bind_closure(b)(80, "0.0.0.0"))._bind_tups
        assert(out == {1: ('0.0.0.0', 80), 2: ('0.0.0.0', 80)})

        # This is where things tend to go badly.
        b.af = IP6

        out = (await bind_closure(b)(80, "::1"))._bind_tups


        out = (await bind_closure(b)(80, "fe80::6c00:b217:18ca:e365"))._bind_tups
        #assert(out[1][2] > 0 or out[2][2] > 0)


        #out = (await bind_closure(b)(80, ""))._bind_tups
        
    async def test_bind(self):
        i = await Interface().start_local()
        af = i.stack if i.stack != DUEL_STACK else IP4
        b = Bind(i, af)

    async def test_ip_val_v6_bind_types(self):
        i = await Interface().start_local()
        try:
            # Test global tuples set.
            ip = ip_norm("2402:1f00:8101:83f::1")
            b = Bind(i, IP6, ips=ip)
            await b.bind()
            self.assertEqual(b.ext_bind, ip)
            self.assertEqual(ip_norm(b.bind_tup(flag=EXT_BIND)[0]), ip)
            self.assertEqual(ip_norm(b.ext_bind), ip)

            # Test link local tuples set.
            ip = i.route(IP6).nic()
            b = Bind(i, IP6, ips=ip)
            await b.bind()
            self.assertEqual(b.nic_bind, ip)
            self.assertEqual(ip_norm(b.bind_tup(flag=NIC_BIND)[0]), ip)
            self.assertEqual(ip_norm(b.nic_bind), ip)
            s = await socket_factory(b)
            self.assertTrue(isinstance(s, socket.socket))
            s.close()
        except:
            return

    async def test_ip_val_v4_bind_types(self):
        i = await Interface().start_local()

        # Test nic bind occurs.
        tests = ["192.168.0.1", "8.8.8.8"]
        for ip in tests:
            b = Bind(i, IP4, ips=ip)
            self.assertEqual(b.nic_bind, ip)
            await b.bind()
            self.assertEqual(b.bind_tup(flag=NIC_BIND)[0], ip)

    async def test_route_v4_bind_types(self):
        i = await Interface().start_local()
        r = i.route(IP4)
        b = await r.bind()
        tup = b.bind_tup(flag=NIC_BIND)
        self.assertTrue(tup[0])

    async def test_route_v6_bind_types(self):
        i = await Interface().start_local()

        try:
            af = IP6
            r = await i.route_test(af)

            # Patch EXT to return a pub IP without interface being resolved.
            r.ext = lambda: "2402:1f00:8101:083f:0000:0000:0000:0001"
            b = await r.bind()
            tup = b.bind_tup(flag=EXT_BIND)
            self.assertTrue(tup[0])
        except:
            return

    # TODO: netifaces is pulling invalid net masks for some IPs?
    async def test_bind_assumptions(self):
        ip = "139.99.209.1"
        #socket.socket(IP4, TCP)


    async def test_bind_start_v4_all_addr(self):
        af = IP4
        try:
            i = await Interface(af).start_local()
        except:
            # If not supported.
            return

        route = await i.route(af).bind(13453, "*")
        bind_tup = ("0.0.0.0", 13453)
        expected_tups = {
            EXT_BIND: bind_tup,
            NIC_BIND: bind_tup
        }
        self.assertEqual(route._bind_tups, expected_tups)
        s = await socket_factory(route)
        self.assertTrue(s is not None)
        if s is not None:
            s.close()

    async def test_bind_start_v6_all_addr(self):
        i = await Interface().start_local()
        try:
            af = IP6
            route = await i.route(af).bind(13453, "*")
            bind_tup = (
                "::",
                13453,

                # NIC no and scope ID stuff from getaddrinfo.
                route._bind_tups[EXT_BIND][2],
                route._bind_tups[EXT_BIND][3]
            )
            expected_tups = {
                EXT_BIND: bind_tup,
                NIC_BIND: bind_tup
            }
            self.assertEqual(route._bind_tups, expected_tups)
            s = await socket_factory(route)
            self.assertTrue(s is not None)
            if s is not None:
                s.close()
        except:
            return

# TODO loopback tests.

if __name__ == '__main__':
    main()


from p2pd.test_init import *
from p2pd.settings import *
from p2pd.signaling import SignalMock
from p2pd.utils import rand_plain, to_s

class TestSignaling(unittest.IsolatedAsyncioTestCase):
    async def test_node_signaling(self):
        if not P2PD_TEST_INFRASTRUCTURE:
            return

        # Channel that the test node subs to.
        i = await Interface().start_local()
        dest_id = "p2pd_test_node"
        msg = "msggg"
        f = asyncio.Future()

        # Receive a message from our MQTT channel (node ID.)
        async def proc_msg(msg, signal_pipe):
            f.set_result(to_s(msg))

        # Start the MQTT client.
        node_id = node_name(b"node_c", i)
        client = None
        for i in range(0, 3):
            client = SignalMock(node_id, proc_msg, mqtt_server=MQTT_SERVERS[i])
            await client.start()

            # Use basic echo protocol to test signaling works.
            await client.echo(msg, dest_id)

            try:
                out = await asyncio.wait_for(
                    f,
                    8
                )
                break
            except Exception:
                what_exception()
                continue

        # Check client set.
        if client is None:
            raise Exception("Couldn't connect to an mqtt server.")


        # Check results and cleanup.
        self.assertTrue(msg in out)
        await client.close()

if __name__ == '__main__':
    main()


from p2pd import *

servers = [


]

u = []
[u.append(i) for i in servers if i not in u]
servers = u


async def main(servers):
    i = await Interface().start_local()
    loop = asyncio.get_event_loop()

    #servers = [["jump.chat", 3478]]
    #servers = [["webrtc.free-solutions.org", 3478], ]

    """
    v6 = []
    for server in servers:
        server = [server["host"], server["port"]]
        x = await loop.getaddrinfo(host=server[0], port=server[1])
        for d in x:
            if d[0] == IP6:
                v6.append(server)

    print(v6)

    return
    """

    # Open servers list and skip done results.
    sfp = open("servers.txt", "r")
    got = eval("[" + sfp.read() + "]")
    print(f"Skipping {len(got)} servers already processed.")
    x = [s for s in servers if s not in got]
    servers = x
    sfp.close()
    sfp = open("servers.txt", "a")

    # Open filtered results.
    tfp = open("tcp.txt", "a")
    ufp = open("udp.txt", "a")
    for server in servers:
        print(f"trying {server}")
        async def worker():
            # UDP must be used for NAT test.
            s = STUNClient(interface=i)
            s.proto = UDP
            nat_type = await s.get_nat_type(servers=[server])
            if nat_type == 3:
                ufp.write(f"{server}\r\n")
                print(f"udp {server}")

            # We also need a server that supports TCP.
            # TCP is needed for the mapping test.
            s.proto = TCP
            m = await s.get_mapping(servers=[server], proto=TCP)
            if m[2] is not None:
                tfp.write(f"{server}\r\n")
                print(f"tcp {server}")

            return None

        sfp.write(f"{server},\r\n")
        await worker()
        #break

    sfp.close()
    tfp.close()
    ufp.close()

async_test(main, args=[servers])


import socket
from p2pd.net import IP4, IP6
from p2pd.route import rp_from_fixed

# Only used for tests.
P2PD_NET_V4_IP = "139.99.209.63"
P2PD_NET_V6_IP = "2402:1f00:8101:83f::1"
P2PD_NET_IPS = {
    IP4: P2PD_NET_V4_IP,
    IP6: P2PD_NET_V6_IP
}

P2PD_NET_V4_FIXED_ROUTES = [
    # Route 0.
    [
        # NIC IPs.
        [
            [
                "139.99.209.63"
            ],
            [
                "139.99.250.35"
            ]
        ],

        # EXT IPs.
        [
            [
                "139.99.209.63"
            ]
        ]
    ]
]

P2PD_NET_V6_FIXED_ROUTES = [
    # Route 0.
    [
        # NIC IPs.
        [
            [
                "fe80::ae1f:6bff:fe94:531a"
            ],
            [
                "2402:1f00:8101:83f::1"
            ]
        ],

        # EXT IPs.
        [
            [
                "2402:1f00:8101:83f::1"
            ]
        ]
    ]
]

def use_fixed_rp(interface):
    rp = None
    if socket.gethostname() == "p2pd.net":
        rp = {}
        rp[IP4] = rp_from_fixed(
            P2PD_NET_V4_FIXED_ROUTES,
            interface,
            IP4
        )

        rp[IP6] = rp_from_fixed(
            P2PD_NET_V6_FIXED_ROUTES,
            interface,
            IP6
        )

    return rp



import ipaddress
import copy
from functools import total_ordering
from .net import *

CIDR_WAN = 1000

class IPRangeIter():
    def __init__(self, ipr, reverse=False):
        self.ipr = ipr
        self.reverse = reverse
        self.host_p = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.host_p >= self.ipr.host_no:
            raise StopIteration

        if self.reverse == False:
            ipa_ip = self.ipr[self.host_p]
        else:
            ipa_ip = self.ipr[(len(self.ipr) - 1) - self.host_p]

        self.host_p += 1
        return ipa_ip

"""
Accepts str, int, bytes for IP and netmask.
Can be converted to str, int, or bytes.
Iterable and sliceable -- returns ip_addr objs.
"""
@total_ordering
class IPRange():
    def __init__(self, ip, netmask=None, cidr=CIDR_WAN):
        # Prefer netmask over cidr.
        if netmask != None and cidr != None:
            cidr = None

        # Sanity check.
        assert(netmask is not None or cidr is not None)
        assert(ip != netmask)

        # Norm net mask -- remove /n, %iface, and/or explode.
        if isinstance(netmask, str):
            self.netmask = ip_norm(netmask)
        else:
            if netmask is None:
                self.netmask = None
            else:
                self.netmask = netmask

        # Is this IP4 or IP6 -- check for ambiguity.
        self.af = None
        if isinstance(ip, int):
            if ip < (2 ** 31):
                if netmask == None:
                    raise Exception("Ambiguous ip int AF.")
                else:
                    ipa_netmask = ipaddress.ip_address(netmask)
                    self.af = v_to_af(ipa_netmask.version)

        # Norm IP -- remove /n, %iface, and/or explode.
        if isinstance(ip, str):
            self.ip = ip_norm(ip)
        else:
            self.ip = ip

        # Use specific AF.
        if self.af is not None:
            if self.af == IP4:
                self.ipa_ip = ipaddress.IPv4Address(self.ip)
            if self.af == IP6:
                self.ipa_ip = ipaddress.IPv6Address(self.ip)
        else:
            self.ipa_ip = ipaddress.ip_address(self.ip)
            self.af = v_to_af(self.ipa_ip.version)

        # Set netmask from cidr if cidr set.
        if cidr is not None:
            if cidr:
                if cidr == CIDR_WAN:
                    cidr = max_cidr(self.af)

                self.netmask = cidr_to_netmask(cidr, self.af)

            self.cidr = cidr
        else:
            # Convert netmask to CIDR with fast binary operations.
            if netmask is not None:
                ipa_netmask = ipaddress.ip_address(self.netmask)
                self.cidr = hamming_weight(int(ipa_netmask))

        # Blank network portion.
        if not self.cidr:
            if self.af == IP4:
                self.netmask = ZERO_NETMASK_IP4
            else:
                self.netmask = ZERO_NETMASK_IP6

        # Parse IP information.
        max_host_bit_len = max_cidr(self.af)
        assert(self.cidr <= max_host_bit_len)
        host_bit_len = max_host_bit_len - self.cidr

        # IP is network portion + host portion.
        self.i_ip = int(self.ipa_ip)
        if host_bit_len:
            # If a range is specified and host bits are set.
            # Get rid of them.
            self.i_host = get_bits(self.i_ip, l=host_bit_len)
            self.i_ip -= self.i_host
            self.host_no = 1
        else:
            # If CIDR was set to the length of the IP
            # then there will be no 'host bits'.
            self.i_host = 0
            self.host_no = 1
            self.i_nw = self.i_ip

        # Blank host portion means this is a range of IPs.
        # That is - it is a network.
        if host_bit_len:
            self.i_nw = self.i_ip
            if host_bit_len != max_host_bit_len:
                self.host_no = (2 ** host_bit_len) - 1

        # IP may have a blank host portion but the set bits
        # still seem to provide enough info for this to work.
        self.is_private = self.ipa_ip.is_private
        self.is_public = not self.is_private
        if not self.i_ip:
            self.is_public = True
            self.is_private = False
        if self.ip in BLACK_HOLE_IPS.values():
            self.is_public = True
            self.is_private = False

        # Used for range comparisons.
        if self.cidr == max_cidr(self.af):
            self.r = [self.i_nw, self.i_nw]
        else:
            self.r = [self.i_nw, self.i_nw + self.host_no]

        assert(self.host_no)

    def len(self):
        return self.host_no

    def ip_f(self, n):
        if self.af == IP4:
            return ipaddress.IPv4Address(n)
        if self.af == IP6:
            return ipaddress.IPv6Address(n)

    def to_dict(self):
        return {
            "ip": self.ip,
            "cidr": self.cidr,
            "af": int(self.af)
        }

    @staticmethod
    def from_dict(d):
        return IPRange(ip=d["ip"], cidr=d["cidr"])

    # Pickle.
    def __getstate__(self):
        return self.to_dict()

    # Unpickle.
    def __setstate__(self, state):
        o = self.from_dict(state)
        self.__dict__ = o.__dict__

    def __deepcopy__(self, memo):
        ip = self.ip
        netmask = self.netmask
        params = (ip, netmask, copy.deepcopy(self.cidr))
        return IPRange(*params)

    def __int__(self):
        return self.i_nw + self.i_host

    def __bytes__(self):
        return i_to_b(int(self))

    def __len__(self):
        return self.host_no

    def __iter__(self):
        return IPRangeIter(self)

    def __reversed__(self):
        return IPRangeIter(self, reverse=True)

    def get_value(self, i):
        """
        Using modulus here means that if ever the left-hand host no
        expression is negative then it will wrap back around the
        number of hosts in the subnet.
        
        This means that negative indexes will work to index
        the subnet. It also provides a safe-guard that its
        impossible to exceed the number of hosts in the subnet
        when indexing it. The extra +1 is added to the host_no
        because we want to be able to include the host_number
        itself in the range.

        The or 1 part is added because when i_nw and i_host are
        set from an IP that's a range: it will have a blank host
        portion. Valid hosts need to start counting from 1. Hence
        we set it to 1 if its 0; otherwise use the existing value.
        """

        # Single WAN IP. One host. Do nothing.
        if self.cidr == max_cidr(self.af):
            return self.ip_f(self.i_nw)
        else:
            # Code to wrap around a subnet.
            if i < 0:
                # Use negative indexing to wrap around host_no.
                offset = i
            else:
                # Start counting at 1.
                offset = i + 1

        # Add network with blank host section to host number.
        # Start counting at zero if host isn't already set.
        i_host = (offset % (self.host_no + 1)) or 1
        return self.ip_f(self.i_nw + i_host)

    def __add__(self, n):
        if isinstance(n, IPRange):
            return self[n.i_host]

        if isinstance(n, int):
            return self[n]

        raise NotImplemented("Add not implemented for that type.")

    def __radd__(self, n):
        return self + n

    def __sub__(self, n):
        if isinstance(n, IPRange):
            return self[-n.i_host]

        if isinstance(n, int):
            return self[-n]

        raise NotImplemented("Sub not implemented for that type.")

    def __rsub__(self, n):
        return self - n

    def _convert_other(self, other):
        if isinstance(other, (int, bytes, str)):
            ipa = ipaddress.ip_address(other)
            return IPRange(ipa, cidr=CIDR_WAN)
        elif isinstance(other, IPRange):
            return other
        elif isinstance(other, IPA_TYPES):
            return IPRange(other, cidr=CIDR_WAN)
        else:
            raise NotImplemented("Compare not implemented for type.")

    def __eq__(self, other):
        other = self._convert_other(other)
        return range_intersects(self.r, other.r)

    def __lt__(self, other):
        other = self._convert_other(other)

        # Compare highest values in range.
        return self.r[1] < other.r[1]

    def __contains__(self, item):
        return self == item

    def __getitem__(self, key):
        if isinstance(key, slice):
            start, stop, step = key.indices(len(self))
            return [self[i] for i in range(start, stop, step)]
        elif isinstance(key, int):
            return self.get_value(key)
        elif isinstance(key, tuple):
            return [self.get_value(x) for x in key]
        else:
            raise TypeError('Invalid argument type: {}'.format(type(key)))

    def __repr__(self):
        return f"{str(self)}"

    # Get an IPAddress obj at start of range.
    # Convert to a string.
    def __str__(self):
        return ipr_norm(self)

    def __hash__(self):
        return hash(str(self))

def ipr_in_interfaces(needle_ipr, if_list, mode=IP_PUBLIC):
    af = needle_ipr.af
    for interface in if_list:
        routes = interface.rp[af].routes
        for route in routes:
            if mode == IP_PUBLIC:
                search_list = route.ext_ips
            if mode == IP_PRIVATE:
                search_list = route.nic_ips

            for hey_ipr in search_list:
                if needle_ipr in hey_ipr:
                    return True

    return False

def ipr_norm(ipr):
    return ip_norm(str(ipr[0]))

if __name__ == "__main__": # pragma: no cover
    # Blank host = range.
    x = IPRange("192.168.1.0", "255.255.255.0")
    assert(str(x[0]) == "192.168.1.1")
    assert(str(x[1]) == "192.168.1.2")
    assert(str(x[-1]) == "192.168.1.255")
    assert(str(x[-2]) == "192.168.1.254")
    assert(x.host_no == 255)

    # Not blank host = single host. Not a range.
    y = IPRange("192.168.1.179", "255.255.255.0")
    assert(str(y[0]) == "192.168.1.179")
    assert(str(y[1]) == "192.168.1.179")
    assert(str(y[-1]) == "192.168.1.179")
    assert(str(y[-2]) == "192.168.1.179")
    #assert(str(x[0]) == "192.168.1.1")
    assert(y.host_no == 1)

    # Single host (with full net mask). Also not a range.
    z = IPRange("7.7.7.7", "255.255.255.255")
    assert(str(z[0]) == "7.7.7.7")
    assert(str(z[15]) == "7.7.7.7")
    assert(str(z[-15]) == "7.7.7.7")
    assert(z.host_no == 1)

    a = IPRange("7.7.7.7", "255.255.255.255")
    b = IPRange("7.7.7.7", "255.255.255.255")
    c = IPRange("7.7.7.8", "255.255.255.255")
    d = IPRange("192.168.1.1", "255.255.255.0")
    e = IPRange("192.168.1.0", "255.255.255.0")
    f = IPRange("192.169.0.0", "255.255.0.0")
    g = IPRange("192.168.2.1", "255.255.255.0")
    h = IPRange("192.168.1.20", "255.255.255.0")
    assert(a == b) # Same IP
    assert(b < c) # CMP single ip values
    assert(a != c) # Not same IP
    assert(d == e) # Check if IP in a range.
    assert(f != e) # Compare two ranges for intersection.
    assert(b < e) # Compare end value of ranges.
    assert(e > b)

    #print(len(f))
    #print(len(e))
    assert(f > e) # Range compare is based on host no, not ip value

    l = [a, c, e]
    assert(d in l)
    assert(b in l)
    assert(g not in l)
    assert(h in l)
    x = IPRange("fe80::9acb:c90e:7bf6:a093%enp3s0", "ffff:ffff:ffff:ffff::/64")
    assert(x.cidr == 64)


    exit(0)


    x = IPRange("192.168.0.0", "255.255.255.0")
    y = IPRange("192.169.0.20", "255.255.255.0")
    print(x == y)
    print(x + 100)
    exit(0)


    print(repr(x))

    print("x[0")
    print(x[600])

    exit(0)
    print(x.i_ip)
    print(x.i_host)
    print(x.i_nw)


    """
    for ipa in x:
        print(ipa)
    """

    #print(x[1: 2: 3]).indices()

    a = x[0:2]
    print(a)



from p2pd import *

async def example():
    #
    # Load default interface.
    i = await Interface().start()
    #
    # Get first supported address family.
    # E.g. if the NIC only supports IPv4 this will == [AF_INET].
    # If it supports both families it will == [AF_INET, AF_INET6].
    # If no AF for route and address specified = use i.supported()[0].
    af = i.supported()[0]
    #
    # Get a route to use for sockets.
    # This will give you a copy of the first route for that address family.
    # Routes belong to an interface and include a reference to it.
    route = await i.route(af).bind()
    #
    # Lookup Google.com's IP address -- specify a specific address family.
    # Most websites support IPv4 but not always IPv6.
    # Interface is needed to resolve some specialty edge-cases.
    dest = await Address("8.8.8.8", 53, route).res()
    #
    # Now open a TCP connection to that the destination.
    pipe = await pipe_open(TCP, route, dest)
    #
    # Send it a malformed HTTP request.
    buf = b"Test\r\n\r\n"
    await pipe.send(buf)
    #
    # Wait for any message response.
    out = await pipe.recv(timeout=3)
    #
    # Cleanup.
    await pipe.close()

if __name__ == '__main__':
    async_test(example)


"""
FIN and RST on close:

Something important to note about TCP hole punching: suppose your NAT type is one that reuses mappings under certain conditions. You go ahead and use STUN to setup the mapping and then retrieve what it is. At this point if you were to go ahead and close() the socket used for the STUN transactions it would send out a 'FIN' packet. 

Now, if the router saw that FIN it may decide to close the port mapping. We would hope that multiple outbound connections with the same local tuple to different destinations would mean that a single FIN wouldn't cause the NAT to disregard other packets. But in any cause the STUN connection ought to stay open at least until the TCP punching has been completed.

Synchronicity:

TCP hole punching between two peers heavily depends on synchronized actions.
If a peer sends a SYN too early then they risk having the other side's NAT
send back an RST which will cause the connection to be closed even if the
connection later 'succeeds.' Therefore, writing one of the peers to
'start sooner' than an agreed upon time simply makes their connection unstable.
The correct way to do TCP hole punching is to start at the same time but
synchronicity between networked devices is a notoriously hard problem to solve.

Using the NTP protocol it can be off by between 1 - 50 ms which is significant
enough to matter. The algorithm tries to compensate by sampling and using
statistical methods to remove outliers. The original approach for this was
taken from the Gnuttella code (written in C) and ported to Python.

Wait times:

The protocol is meant to take up to N seconds to complete. Where N is enough
time to receive initial mappings and / or updated mappings + a buffer for
synchronized startup. There is also a limit on the whole process imposed
by the 'pipe waiter' code that waits for a pipe to be returned. If the
limit is too short then you may see connections succeed in the punching
code but then be tore down from the timeout cleanup code. I wanted to
make a note of this here. I am trying to minimize how much time is spent
sleeping so that the process is faster but I am still tweaking this.

Sharing sockets between processes:

The TCP punch module uses Python's ProcessPoolExecutor to spawn child processes
that do the punching operations. Using this design has several benefits. It
means that the timing between peers is slightly more accurate as the main process
is not interrupting the connection code. In reality though: the operating system
still controls scheduling of processes (without using special code to pin the
processes to certain cores.)

Another benefit is that parts of the code which are very 'busy' such as the local
punching algorithm (which essentially just spams connections as much as it can)
will not impact the performance of the main application. There is a consequence
to this design though. If a process exits then presumably the socket descriptors
opened from the process will become unusable. Fortunately, Python's process pool
spins up a list of processes (by default set to the core number) and reuses them.
That means that the sockets should remain valid through the software's lifetime.

Event loops:

On Linux, Mac OS X, BSD... The default event loop is the selector event loop.
On Windows the default event loop is the proactor event loop. When it comes
to running commands and making 'pipes' to processes you need to use the
proactor event loop on Windows. But for TCP hole punching to work the
selector event loop is the only one that seems to make the code work.

When running async code for the first time in an 'executor' / new process
it will need to create a new event loop. Normally this would have quite
a delay. In order to make async code run fast I pre-initialize all executors
with a call to create an event looop. So when async code is run in them
there is no startup penalty. This is important for timing-based code.

Testing:

When it comes to testing punching on the same machine it is necessary to run the punching code in separate processes (with their own event loops.) Otherwise,
they will interfere with each other and the syns won't cross in time. Ideally
each process should be running on its own core with a high priority. But this
is hard to guarantee in practice.

Edge-case: making another connection to the same STUN server, from the same local endpoint due to another TCP punch occuring.
"""

import socket
import multiprocessing
import warnings
import asyncio
from concurrent.futures import ProcessPoolExecutor
from .nat import *
from .ip_range import *
from .stun_client import *
from .clock_skew import *
from .base_stream import *
from .interface import *

#nest_asyncio.apply()
asyncio.set_event_loop_policy(SelectorEventPolicy())

# Config #######################################################
# Number of seconds in the future from an NTP time
# for hole punching to occur.
NTP_MEET_STEP = 3

################################################################
TCP_PUNCH_MAP_NO = 5
TCP_PUNCH_SEND_INITIAL_MAPPINGS = 1
TCP_PUNCH_RECV_INITIAL_MAPPINGS = 2
TCP_PUNCH_UPDATE_RECIPIENT_MAPPINGS = 3
TCP_PUNCH_IN_PROGRESS = 4
TCP_PUNCH_SUCCESS = 5
TCP_PUNCH_FAILURE = 6
PUNCH_INITIATOR = 1
PUNCH_RECIPIENT = 2

# Fine tune various network settings.
PUNCH_CONF = dict_child({
    # Reuse address tuple for bind() socket call.
    "reuse_addr": True,

    # Return the sock instead of the base proto.
    "sock_only": True,

    # Disable closing sock on error.
    "no_close": True,

    # Ref to async event loop func.
    "loop": lambda: selector_event_loop()
}, NET_CONF)

# Just merges two specific lists into slightly different format.
def map_info_to_their_maps(map_info):
    their_maps = []
    for i in range(0, len(map_info["remote"])):
        their_map = [
            map_info["remote"][i],
            map_info["reply"][i],
            map_info["local"][i]
        ]

        their_maps.append(their_map)

    return their_maps

# Started in a new process.
def proc_do_punching(args):
    # Create new event loop and run coroutine in it.
    """
    On Windows it seems like using the default 'proactor event loop'
    prevents the TCP hole punching code from working. It seems that
    manually setting the event loop to use SelectorEventLoop
    fixes the issue. However, this may mean breaking some of
    my command execution code on Windows -- test this.
    """
    asyncio.set_event_loop_policy(SelectorEventPolicy())
    side = args.pop()
    sq = args.pop()
    q = args.pop()

    loop = asyncio.get_event_loop()
    f = asyncio.ensure_future(
        async_wrap_errors(
            TCPPunch.do_punching(side, *args, sq)
        ),
        loop=loop
    )



    f.add_done_callback(lambda t: q.put(t.result()))
    loop.run_until_complete(f)
    #return f.result()


state_info_index = lambda x, y: repr(x) + "@" + str(y)
class TCPPunch():
    def __init__(self, interface, if_list, sys_clock=None, executors=None, queue_manager=None, conf=PUNCH_CONF):
        self.conf = PUNCH_CONF
        self.if_list = if_list
        self.interface = interface
        self.sys_clock = sys_clock
        self.executors = executors
        self.queue_manager = queue_manager
        if self.interface.nat["type"] in BLOCKING_NATS:
            raise Exception("Peer NAT type is unreachable.")

        self.active_no = 0

        # [node_id][pipe_id] = state
        self.state = {}

        # Keep a list of socks used to open TCP mappings
        # to STUN server. If a sock goes out of scope then
        # Pythons garbage collector reaps it and will call
        # close, we want to save it until after a hole punch.
        self.stun_socks = []
        if queue_manager is not None:
            self.sock_queue = queue_manager.Queue()
        else:
            self.sock_queue = None

        # Simple garbage collector for state.
        self.do_state_cleanup = True
        self.tcp_punch_stopped = asyncio.Event()
        self.stop_running = asyncio.Event()
        self.cleanup_task = asyncio.create_task(
            async_wrap_errors(
                self.cleanup_by_timeout()
            )
        )

    async def close(self):
        # Already closed.
        if self.stop_running.is_set():
            return

        self.stop_running.set()
        await self.tcp_punch_stopped.wait()

    def get_punch_mode(self, dest_addr):
        # Loopback so use local punching.
        if dest_addr.is_loopback:
            return TCP_PUNCH_SELF

        # External IP that belongs to ourself.
        # Use local punching.
        needle_ipr = IPRange(dest_addr.tup[0])
        if needle_ipr.is_public:
            # Check all external addresses for routes.
            if ipr_in_interfaces(needle_ipr, self.if_list, mode=IP_PUBLIC):
                return TCP_PUNCH_SELF

            # It's possible to use public IPs for NIC addresses.
            # Check whether it matches any in our routes.
            if ipr_in_interfaces(needle_ipr, self.if_list, mode=IP_PRIVATE):
                return TCP_PUNCH_SELF

        # Private LAN IP so use local punching.
        if dest_addr.is_private:
            # Is this our own NIC address?
            if ipr_in_interfaces(needle_ipr, self.if_list, mode=IP_PRIVATE):
                return TCP_PUNCH_SELF
            else:
                return TCP_PUNCH_LAN

        # Otherwise use remote punching.
        return TCP_PUNCH_REMOTE

    def get_ntp_meet_time(self):
        if self.sys_clock is None:
            return 0

        return self.sys_clock.time() + Dec(NTP_MEET_STEP)

    def get_state_info(self, node_id, pipe_id):
        if node_id not in self.state:
            return None

        if pipe_id not in self.state[node_id]:
            return None

        return self.state[node_id][pipe_id]

    def set_state(self, node_id, pipe_id, state, data=None):
        # Update session counter.
        self.active_no += 1

        # State data format.
        info = {
            "node_id": node_id,
            "pipe_id": pipe_id,
            "state": state,
            "data": data,
            "timestamp": timestamp()
        }

        # Create session table for node.
        if node_id not in self.state:
            self.state[node_id] = {}

        # Set state for node.
        self.state[node_id][pipe_id] = info

    """
    Handles closing any open STUN handles.

    Removes a specific state dictionary. Decrease session counters.
    May remove session counter when zero.
    """
    async def cleanup_state(self, node_id, pipe_id):
        # Does the state exist?
        state_info = self.get_state_info(node_id, pipe_id)
        if state_info is None:
            return 0

        # Close any open STUN sockets.
        if "data" in state_info:
            data = state_info["data"]
            if "lmaps" in data:
                for sock in data["lmaps"]["stun_socks"]:
                    if sock is not None:
                        await sock.close()

        # Delete dest@session = session.
        del self.state[node_id][pipe_id]

        # If it was the last pipe delete the whole struct.
        if not len(self.state[node_id]):
            del self.state[node_id]

        if self.active_no:
            self.active_no -= 1

        return 1
        
    """
    The TCP hole punching process builds up connection state.
    It's important that old state gets deleted to avoid filling up memory.
    This is a simple coroutine that fires up every 10 mins.
    It cleans up all stale state and then sleeps.
    """
    async def cleanup_by_timeout(self, max_elapsed=5 * 60):
        while not self.stop_running.is_set():
            now = timestamp()
            async def do_cleanup(state_info):
                elapsed = now - state_info["timestamp"]
                if elapsed >= max_elapsed:
                    await self.cleanup_state(
                        state_info["node_id"],
                        state_info["pipe_id"]
                    )

            # Check for timeouts every ten mins.
            try:
                await asyncio.wait_for(
                    self.stop_running.wait(),
                    max_elapsed
                )
            except asyncio.TimeoutError:
                pass

            # Check timeout for all active sessions, for all dests.
            """
            The cleanup isn't done in the loop because changing the
            size of state while iterating over it with keys causes
            an exception. Instead the cleanup tasks are saved and
            then run sequentially.
            """
            tasks = []
            for node_id in self.state.keys():
                for pipe_id in self.state[node_id].keys():
                    state_info = self.get_state_info(node_id, pipe_id)
                    tasks.append(do_cleanup(state_info))

            # Cleanup all state structs.
            for task in tasks:
                await task

        self.tcp_punch_stopped.set()

    # Initiator: Step 1 -- send Initiators predicted mappings to Recipient.
    async def proto_send_initial_mappings(self, dest_addr, dest_nat, dest_node_id, pipe_id, stun_client, process_replies=None, con_list=None, mode=TCP_PUNCH_REMOTE):
        assert(stun_client.interface == self.interface)

        # Log warning if dest_addr is our ext address.
        af = af_from_ip_s(dest_addr)
        ext = self.interface.route(af).ext()
        if dest_addr == ext:
            log("> TCP punch warning: step 1 dest matches our dest.")

        # Session id is simple len of existing sessions.
        map_info = await get_nat_predictions(
            mode,
            stun_client,
            self.interface.nat,
            dest_nat
        )

        # Sanity check.
        nat_check_for_low_ports(map_info)

        """
        The NAT prediction code tries to choose ports in a way
        where both parties trying to connect can use the same
        remote port. The reason for this is if host A knows that
        host B will also try use the same ports -- then it may
        be able to connect to host B even if host B can't get
        back a message to host A with its remote mappings.

        It's not always possible to do this depending on the
        NAT types involved. But it's done when it's possible.
        Just another way I wanted to make this code more robust.
        """

        # Triggered on receiving updated mappings (if any.)
        update_event = asyncio.Event()
        
        # Save session data.
        our_maps = map_info_to_their_maps(map_info)
        ntp_meet = self.get_ntp_meet_time()
        data = {
            # Defines the algorithm used for punching.
            "mode": mode,

            # Event for step 3.
            "update_event": update_event,

            # The NAT predictions for ourself live here.
            "lmaps": map_info,

            # By default assume their maps will be our maps.
            # Overwritten if we receive a reply.
            "rmaps": our_maps,

            # Their NAT struct.
            "their_nat": dest_nat,

            # Their dest.
            "their_dest": dest_addr,

            # Their node id.
            "their_node_id": dest_node_id,

            # Pipe ID used for session.
            "pipe_id": pipe_id,

            # These are all optional:
            ##########################

            # Protocol handler that runs for a punched con.
            "process_replies": process_replies,

            # Used for cleanup / insert to shared state.
            "con_list": con_list,

            # Future time for hole punching to occur.
            "ntp_meet": ntp_meet
        }

        # First state progression -- no need to check existing.
        self.active_no += 1
        self.set_state(
            dest_node_id,
            pipe_id,
            TCP_PUNCH_SEND_INITIAL_MAPPINGS,
            data
        )

        return our_maps, ntp_meet, update_event

    # Recipient: Step 2 -- get Initiators mappings. Generate our own.
    async def proto_recv_initial_mappings(self, recv_addr, recv_nat, recv_node_id, pipe_id, their_maps, stun_client, ntp_meet=0, process_replies=None, con_list=None, mode=TCP_PUNCH_REMOTE):
        assert(stun_client.interface == self.interface)

        # Invalid len.
        if not is_valid_rmaps(their_maps):
            log("> tpc punch recv: invalid rmaps")
            return None
        their_maps = rmaps_strip_duplicates(their_maps)
        if not valid_mappings_len(their_maps):
            log("> tpc punch recv: invalid rmaps len")
            return None

        # Log warning if dest_addr is our ext address.
        af = af_from_ip_s(recv_addr)
        ext = self.interface.route(af).ext()
        if recv_addr == ext:
            log("> TCP punch warning: step 2 recv matches our dest.")

        # Should be no other state progressions.
        existing = self.get_state_info(recv_node_id, pipe_id)
        if existing is not None:
            log("> tpc punch recv: state exists")
            return None

        # Get NAT predictions relative to theirs.
        # If required (and if we can) - change our bind port.
        map_info = await get_nat_predictions(
            mode,
            stun_client,
            self.interface.nat,
            recv_nat,
            their_maps
        )

        # Sanity check.
        nat_check_for_low_ports(map_info)

        # Expect them to use our reply port if set.
        """
        It's possible that the Initiator has already chosen a remote
        port that matches what our reply port will be. In which case
        this code will have no effect. The code intelligently
        selects ports to use based on its partners NAT type.

        Consequently, the Initiator doesn't always need a reply
        from the Recipient. For example: if you lookup a NAT mapping
        via STUN on a 'port restricted NAT' then the NAT will need
        a peer to use the same port as the STUN service for the
        mapping to be used. STUN listens on a common port so this
        may be selected by the Initator in advance.
        """
        if mode != TCP_PUNCH_SELF:
            x = len(map_info["reply"])
            y = len(their_maps)
            for i in range(0, min(x, y)):
                if map_info["reply"][i]:
                    # Overwrite their remote port.
                    their_maps[i][0] = map_info["reply"][i]

        # State data to change.
        data = {
            "mode": mode,
            "lmaps": map_info,
            "rmaps": their_maps,
            "process_replies": process_replies,
            "con_list": con_list,
            "ntp_meet": ntp_meet,
            "their_nat": recv_nat,
            "their_dest": recv_addr,
            "their_node_id": recv_node_id,
            "pipe_id": pipe_id
        }

        # Update state.
        self.active_no += 1
        our_maps = map_info_to_their_maps(map_info)
        self.set_state(
            recv_node_id,
            pipe_id,
            TCP_PUNCH_RECV_INITIAL_MAPPINGS,
            data
        )

        # Sent update won't be important.
        # I just wanted the first two sides to have the same ret.
        sent_update = asyncio.Event()
        return our_maps, ntp_meet, sent_update
        
    """
    Initiator: optional -- may receive Recipients mappings.
    If they do not, then they assume they are using any
    specified reply ports or the same mappings as their own.
    Obviously only possible based on Recipients NAT.
    Such a step is --required-- if attemping to punch self
    as both sides will need to use unique bind ports.
    """
    async def proto_update_recipient_mappings(self, dest_node_id, pipe_id, their_maps, stun_client):
        if stun_client.interface != self.interface:
            raise Exception("Invalid interface.")

        # Invalid len.
        if not is_valid_rmaps(their_maps):
            log("> update map -- invalid rmaps")
            return None
        their_maps = rmaps_strip_duplicates(their_maps)
        if not valid_mappings_len(their_maps):
            log("> update map -- invalid len")
            return None

        # Short-hand references to huge names.
        state_info = self.get_state_info(dest_node_id, pipe_id)
        if state_info is None:
            log("> update map -- state info none")
            return None
        data = state_info["data"]
        lmaps = data["lmaps"]

        # Only possible after initial state update.
        test_no = len(their_maps)
        use_range = nats_intersect_range(self.interface.nat, data["their_nat"], test_no)
        bad_delta = [INDEPENDENT_DELTA, DEPENDENT_DELTA, RANDOM_DELTA]
        if state_info["state"] == TCP_PUNCH_SEND_INITIAL_MAPPINGS:
            # Update our local ports if needed.
            for i in range(0, test_no):
                _, reply_port, _ = their_maps[i]
                if reply_port:
                    # We can satisfy their requirements.
                    if self.interface.nat["delta"]["type"] not in bad_delta:
                        # local, remote, reply, sock.
                        m, _ = await get_single_mapping(
                            data["mode"],
                            their_maps[i],
                            lmaps["last_mapped"],
                            use_range,
                            self.interface.nat,
                            stun_client
                        )

                        # Update our local port.
                        if data["mode"] != TCP_PUNCH_SELF:
                            new_local, _, _, _ = m
                            lmaps["local"][i] = new_local
                            lmaps["remote"][i] = reply_port

                    # Else = they can't connect.
            
            # Trucate our maps if needed.
            data["rmaps"] = their_maps
            lmaps["local"] = lmaps["local"][:test_no]
            lmaps["remote"] = lmaps["remote"][:test_no]
            lmaps["reply"] = lmaps["reply"][:test_no]
            
            # Indicate our recipient mappings were updated.
            self.set_state(
                dest_node_id,
                pipe_id,
                TCP_PUNCH_UPDATE_RECIPIENT_MAPPINGS,
                data
            )

            data["update_event"].set()
            return self.get_state_info(dest_node_id, pipe_id)
        else:
            log("> update map -- invalid state")
            data["update_event"].set()

    # Both sides: step 3.
    """
    I use 'SO_REUSEPORT' to open another connect
    socket and keep the other STUN sockets open.
    """
    @staticmethod
    async def do_punching(side, interface, our_wan, dest_addr, af, local, remote, rmaps, current_ntp, ntp_meet, mode, sock_queue):
        """
        Punching is done in its own process.
        The process returns an open socket and Python
        warns that the socket wasn't closed properly.
        This is the intention and not a bug!
        This code disables that warning.
        """
        warnings.filterwarnings('ignore', message="unclosed", category=ResourceWarning)

        # Init.
        log(f"> TCP punch interface rp pool")
        log_interface_rp(interface)
        is_connected = [False]
        socks = []

        # If NTP meet is defined then wait for it to occur.
        if ntp_meet:
            # NTP meet time is too large -- set to max.
            """
            max_meet_time = current_ntp + NTP_MEET_STEP
            if max_meet_time < ntp_meet:
                ntp_meet = max_meet_time
            """

            # Sleep until the ntp timeframe.
            if current_ntp < ntp_meet:
                remaining_time = float(ntp_meet - current_ntp)
                if remaining_time:
                    log(
                        "> punch waiting for meeting = %s" %
                        (str(remaining_time))
                    )

                    await asyncio.sleep(remaining_time)
            else:
                log("TCP punch behind current meeting time!")

        """
        We can keep trying until success without the mappings
        changing. However, unpredictable nats that use
        delta N will be timing sensitive. If timing info
        is available the protocol should try use that.
        """
        async def delay_con(start_time, ms_delay, local_port, remote_port, dest, sock_timeout, is_connected, loop):
            # Used for making new cons.
            conf = copy.deepcopy(PUNCH_CONF)
            conf["con_timeout"] = sock_timeout 

            """
            Schedule connection to run across time.

            How long it takes the event loop to start a
            routine is outside of our control. This
            code adjusts the delay based on any amount
            of time already passed.
            """
            if ms_delay:
                await asyncio.sleep(ms_delay / 1000)

            # Bind to a specific port and interface.
            route = await interface.route(af).bind(local_port)

            # Open connection -- return only sock.
            #sock = await pipe_open(TCP, dest, route=route, conf=PUNCH_CONF) 
            sock = await socket_factory(route=route, dest_addr=dest, conf=PUNCH_CONF)
            #sock_queue.put_nowait(sock)
            #sock.settimeout(0.1)
            try:
                await loop.sock_connect(
                    sock,
                    dest.tup
                )
            except:
                return None

            # Failure.
            # We strip None responses out.
            if sock is None:
                return None
            else:
                is_connected[0] = True

            """
            At some point I expect errors to be made if a
            local IP + local port is reused for the same
            dest IP + dest port and it successfully gets
            connected. So successive calls might fail.
            """
            return [remote_port, sock]


        """
        Smarter code that is less spammy used for remote
        punching. Won't overwhelm the event loop like the
        above code will. Takes advantage of the massive timeouts
        in the Internet as packets travel across routers.
        Optimized and tested for remote connections.
        """
        async def remote_mode(dest_addr, af, interface, local, remote, rmaps, is_connected):
            #print("mode = remote")
            #print(f"bound on {local}")

            # Con occuring every 
            # second for 3 seconds for
            # all tests at once.
            secs = 6
            ms_spacing = 5
            cons = []
            tasks = []
            start_time = timestamp(1)
            steps = int((secs * 1000) / ms_spacing)
            route = interface.route(af)
            log(f"TCP PUNCH REMOTE local = {local}")
            for i in range(0, len(local)):
                dest = await Address(
                    dest_addr,
                    rmaps[i][0],
                    route
                ).res()

                #print(f"connect to {dest.tup}")
                for sleep_time in range(0, steps):
                    tasks.append(
                        delay_con(
                            # Record start time for logging.
                            start_time,

                            # Wait until ms to do punching.
                            # Punches are split up over time
                            # to increase chances of success.
                            sleep_time * ms_spacing,

                            # Local ports to bind to.
                            local[i],

                            # Used to choose a single con.
                            remote[i],

                            # Destination addr to connect to.
                            dest,

                            # This prevents cons
                            # from being scheduled later
                            # on and exceeding total secs
                            2,
                            is_connected,
                            asyncio.get_event_loop()
                        )
                    )
            
            # Start running tasks.
            all_tasks = asyncio.gather(*tasks)
            outs = await all_tasks
            outs = strip_none(outs)
            return outs

        log("In punch. mode = ")
        log(mode)

        # Use local punching algorithm.
        
        if mode in [TCP_PUNCH_SELF, TCP_PUNCH_LAN]:
            # Updated mapping error occured.
            # Cannot bind to same port as other side on same host.
            if mode == TCP_PUNCH_SELF and remote[0] == rmaps[0][0]:
                log("> Punch updated mapping error for punch self.")
                return None

            # Punch local -- use same algorithm as remote.
            outs = await remote_mode(
                dest_addr=dest_addr,
                af=af,
                interface=interface,
                local=local,
                remote=remote,

                # rmap = [ remote, reply, local ]
                # connect to their local port.
                rmaps=[[rmap[2]] for rmap in rmaps],
                is_connected=is_connected
            )

        # Use remote punching algorithm.
        if mode == TCP_PUNCH_REMOTE:
            outs = await remote_mode(
                dest_addr=dest_addr,
                af=af,
                interface=interface,
                local=local,
                remote=remote,
                rmaps=rmaps,
                is_connected=is_connected
            )

        # Make both sides choose the same socket.
        #print(outs)
        try:
            our_ip_num = ip_str_to_int(our_wan)
            chosen_sock = None
            h_val = 0
            for remote_port, sock in outs:
                their_ip_host, their_r_port = sock.getpeername()
                their_ip_num = ip_str_to_int(
                    their_ip_host
                )

                """
                A TCP connection is defined by a unique tuple of
                src_ip, src_port, dest_ip, dest_port. The purpose
                of this code is to define a single view of the
                'highest' value connection based on the tuple.
                The highest connection will be used in the event
                multiple 'holes' were punched. The clients will
                close the unneeded connections.
                """
                if our_ip_num == their_ip_num:
                    x = sorted([our_ip_num, their_ip_num, remote_port, their_r_port])
                else:
                    if our_ip_num > their_ip_num:
                        x = [our_ip_num, their_ip_num, remote_port, their_r_port]
                    else:
                        x = [their_ip_num, our_ip_num, their_r_port, remote_port]

                # Mix values into a somewhat unique result.
                h = abs(hash(str(x)))
                assert(h > 0)
                if h > h_val:
                    h_val = h
                    chosen_sock = sock
        except Exception as e:
            log_exception()
            log("unknown exception occured")

        # Close all other sockets that aren't needed.
        for _, sock in outs:
            if sock != chosen_sock:
                sock.close()
                pass
        
        # Return sock result.
        if chosen_sock is not None:
            return chosen_sock
        else:
            log("> tcp punch chosen sock is none")
            return None

    def get_punch_args(self, node_id, pipe_id):
        # Get associated state.
        state_info = self.get_state_info(node_id, pipe_id)
        if state_info is None:
            raise Exception("cant find state.")

        dest_addr = state_info["data"]["their_dest"]
        af = af_from_ip_s(dest_addr)
        lmaps = state_info["data"]["lmaps"]
        rmaps = state_info["data"]["rmaps"]
        interface = self.interface
        current_ntp = 0
        if self.sys_clock is not None:
            current_ntp = self.sys_clock.time()
    
        return [
            interface,
            interface.route(af).ext(),
            dest_addr,
            af,
            lmaps["local"],
            lmaps["remote"],
            rmaps,
            current_ntp,
            state_info["data"]["ntp_meet"]
        ]

    """
    This function schedules TCP hole punching to run in
        - a new process
        - a new event loop in said process
        - and handles shared memory of resources
    """
    async def proto_do_punching(self, side, node_id, pipe_id, msg_cb=None):
        # Get associated state.
        state_info = self.get_state_info(node_id, pipe_id)
        if state_info is None:
            log("Cant find state info in proto do punching.")
            return None

        # These arguments are all able to be 'pickled.'
        # This requirement is annoying as hell.
        dest_addr = state_info["data"]["their_dest"]
        if self.queue_manager is not None:
            queue = self.queue_manager.Queue()
        else:
            queue = None
        af = af_from_ip_s(dest_addr)
        route = self.interface.route(af)
        addr = await Address(dest_addr, 0, route).res()
        args = self.get_punch_args(node_id, pipe_id)
        
        # The executor pool is used to start processes.
        loop = asyncio.get_event_loop()
        sock = None
        
        # Multiprocessing pool is enabled.
        if self.executors is not None:
            # Setup args list for executor.
            args.append(state_info["data"]["mode"])
            args.append(queue)
            args.append(self.sock_queue)
            args.append(side)
            
            # Schedule TCP punching in process pool executor.
            loop.run_in_executor(
                self.executors, proc_do_punching, args
            )
            
            # Wait for queue result with timeout.
            for _ in range(0, 30):
                if not queue.empty():
                    break

                await asyncio.sleep(1)

            # Executor pushes back socket to queue.
            try:
                sock = queue.get(timeout=2)
            except Exception:
                log_exception()
                sock = None
        
        # Otherwise do the punching in this event loop.
        if self.executors is None:
            sock = await async_wrap_errors(
                TCPPunch.do_punching(side, *args, state_info["data"]["mode"], self.sock_queue)
            )

        # Wrap returned socket in a pipe.
        pipe = None
        if sock is not None: 
            log(f"> TCP hole made {sock}.")
            route = await self.interface.route(af).bind(
                sock.getsockname()[1]
            )

            pipe = await pipe_open(
                route=route,
                proto=TCP,
                dest=await Address(
                    *sock.getpeername(),
                    route
                ).res(),
                sock=sock,
                msg_cb=msg_cb
            )

        # Cleanup state.
        if self.do_state_cleanup:
            await self.cleanup_state(node_id, pipe_id)

        return pipe
    
async def test_tcp_punch(): # pragma: no cover
    # Use IPv4 for protocol.
    """
    import resource
    resource.setrlimit(resource.RLIMIT_NOFILE, (100000, 100000))
    """
    sabotage = False
    af = socket.AF_INET
    executors = ProcessPoolExecutor()
    m = multiprocessing.Manager()
    loop = asyncio.get_event_loop()

    """
    K, reusing works.
    bind_tup = ("192.168.21.141", 58133)
    sock_a = socket.socket(AF_INET, SOCK_STREAM)
    sock_a.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock_a.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
    sock_a.bind(bind_tup)

    sock_b = socket.socket(AF_INET, SOCK_STREAM)
    sock_b.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock_b.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
    sock_b.bind(bind_tup)

    return
    """

    internode_nat = nat_info(FULL_CONE, delta_info(PRESERV_DELTA, 0))
    starlink_nat = nat_info(RESTRICT_PORT_NAT, delta_info(RANDOM_DELTA, 0), [34000, MAX_PORT])

    # nat type = 5 (restrict nat) - preserving delta.
    # Sends to invalid dest port -- where the hell
    # is this dest port coming from??
    #interfaces = await Interfaces(["enp3s0", "wlp2s0"]).load()
    #internode_out = interfaces.by_name("enp3s0")
    internode_out = await Interface("enp3s0").start()
    internode_out.set_nat(internode_nat)


    #sys_clock = await SysClock().start()
    sys_clock = None # disable for testing

    #print(internode_out.wan_ips)
    #return
    internode_stun_client = STUNClient(internode_out)
    internode_dest = internode_out.route(af).ext()
    
    """
    if sabotage:
        internode_dest = await Address(
            "123.123.233.3",
            0
        ).res()
    """
    print(internode_dest)


    initiator = internode_tcp_punch = TCPPunch(
        internode_out,
        None,
        sys_clock,
        executors,
        m
    )

    # NAT type = restrict port nat, random delta.
    #starlink_out = interfaces.by_name("wlp2s0")
    #print("starlink out wan ips = %s " % (str(starlink_out.wan_ips)))
    starlink_out = await Interface("wlp2s0").start()
    starlink_out.set_nat(starlink_nat)


    starlink_dest = starlink_out.route(af).ext()
    """
    if sabotage:
        starlink_dest = internode_dest
    """

    print(starlink_dest)

    starlink_stun_client = STUNClient(starlink_out)
    #delta = await mapping_test(internode_stun_client)
    #print(delta)

    recipient = starlink_tcp_punch = TCPPunch(
        starlink_out,
        None,
        sys_clock,
        executors,
        m
    )

    """
    local_port = await recipient.brute_force_mapping(40000)
    print(local_port)
    if local_port is None:
        return
    ret = await starlink_stun_client.get_mapping(SOCK_STREAM, source_port=local_port)
    print(ret)

    return
    """

    # Step 1 -- set initial mappings for initiator.
    i_session, i_their_maps, i_ntp_meet = await initiator.proto_send_initial_mappings(
        starlink_dest,
        starlink_nat,
        internode_stun_client
    )

    
    print("I punch details:")
    print(i_session)
    print(i_their_maps)
    print(i_ntp_meet)
    print(initiator.state)
    print()

    # Step 2 -- exchange initiator mappings with recipient.
    # Details would be exchanged over another data channel.
    # In our protocol the SIP client is designed for that.
    r_session, r_their_maps, r_ntp_meet = await recipient.proto_recv_initial_mappings(
        internode_dest,
        internode_nat,
        i_session,
        i_their_maps,
        starlink_stun_client,
        i_ntp_meet
    )

    print("R punch details:")
    print(recipient.state)
    print()

    # Optional step -- update mappings of recipient.
    i_state_info = await initiator.proto_update_recipient_mappings(starlink_dest, i_session, r_their_maps, internode_stun_client)

    print("I state info after update their maps")
    print(initiator.state)
    print()

    print("internode dest = %s" % (internode_dest))
    print("starlink dest = %s" % (starlink_dest))

    results = await asyncio.gather(
        initiator.proto_do_punching(starlink_dest, i_session),
        recipient.proto_do_punching(internode_dest, r_session)
    )

    results = strip_none(results)
    print(results)

    """
    if len(results) >= 2:
        con1, con2 = results
        con1 = await con1.send(b"this is a test")
        data = await con2.recv_n(1024)
        print(data)

    print(results)
    """

    print("Started, now sleeping.")
    while 1:
        await asyncio.sleep(1)

    """
    await asyncio.gather(
        initiator.proto_do_punching(starlink_dest, i_session),
        #recipient.proto_do_punching(internode_dest, r_session)
    )
    """

    return

    # Optional step (but sometimes necessary depending on receiver nat type) -- receiver exchanges
    # their mapping info back to initiator.

    """
    observation: since the stun servers all listen on the
    same port then clients that are port restricted are all
    going to request the same reply ports which is kind of
    not ideal to have 5 identical tests on its partner running
    
    options:
        - less tests? - not ideal
        - more variety in reply ports
            - can use the change port at least
        - research hosts with non-standard bind ports?

    """


    return

    """
    concurrent tcp punches:

    protocol = 

    """



    return

    #delta = await delta_n_test(stun_client)
    #print("Delta = %d" % (delta))



    """
    class TCPPunch():
    def __init__(self, interfaces, stun_client, nat_type=RestricPortNAT, delta=RANDOM_DELTA):
    """

if __name__ == "__main__": # pragma: no cover
    async_test(test_tcp_punch)